@@ -15,7 +15,6 @@
 package app
 
 import (
-	"os"
 	"runtime"
 	"strings"
 	"sync"
@@ -24,6 +23,7 @@ import (
 	"github.com/pkg/errors"
 	log "github.com/sirupsen/logrus"
 
+	"github.com/mendersoftware/mender/app/proxy"
 	"github.com/mendersoftware/mender/client"
 	"github.com/mendersoftware/mender/conf"
 	"github.com/mendersoftware/mender/datastore"
@@ -88,6 +88,7 @@ type AuthManagerRequest struct {
 // AuthManagerResponse stores a response from the Mender authorization manager
 type AuthManagerResponse struct {
 	AuthToken client.AuthToken
+	ServerURL client.ServerURL
 	Event     string
 	Error     error
 }
@@ -137,11 +138,13 @@ type menderAuthManagerService struct {
 	dbus           dbus.DBusAPI
 	dbusConn       dbus.Handle
 	config         *conf.MenderConfig
-	store          store.Store
 	keyStore       *store.Keystore
 	idSrc          device.IdentityDataGetter
-	serverURL      string
+	authToken      client.AuthToken
+	serverURL      client.ServerURL
 	tenantToken    client.AuthToken
+
+	localProxy *proxy.ProxyController
 }
 
 // AuthManagerConfig holds the configuration of the auth manager
@@ -160,24 +163,27 @@ func NewAuthManager(conf AuthManagerConfig) *MenderAuthManager {
 		return nil
 	}
 
-	var api *client.ApiClient
+	httpConfig := client.Config{}
 	if conf.Config != nil {
-		var err error
-		api, err = client.New(conf.Config.GetHttpConfig())
-		if err != nil {
-			return nil
-		}
+		httpConfig = conf.Config.GetHttpConfig()
+
 	}
 
-	// get the first server URL available in the config file
-	serverURL := ""
-	if conf.Config != nil {
-		serverIterator := nextServerIterator(*conf.Config)
-		if serverIterator != nil {
-			if server := serverIterator(); server != nil {
-				serverURL = server.ServerURL
-			}
-		}
+	api, err := client.NewApiClient(httpConfig)
+	if err != nil {
+		return nil
+	}
+
+	tenantToken := client.AuthToken(conf.TenantToken)
+
+	wsDialer, err := client.NewWebsocketDialer(httpConfig)
+	if err != nil {
+		return nil
+	}
+
+	proxy, err := proxy.NewProxyController(api, wsDialer, "", "")
+	if err != nil {
+		log.Errorf("Error creating local proxy: %s", err.Error())
 	}
 
 	mgr := &MenderAuthManager{
@@ -190,11 +196,10 @@ func NewAuthManager(conf AuthManagerConfig) *MenderAuthManager {
 			api:            api,
 			authReq:        client.NewAuth(),
 			config:         conf.Config,
-			store:          conf.AuthDataStore,
 			keyStore:       conf.KeyStore,
 			idSrc:          conf.IdentitySource,
-			tenantToken:    client.AuthToken(conf.TenantToken),
-			serverURL:      serverURL,
+			tenantToken:    tenantToken,
+			localProxy:     proxy,
 		},
 	}
 
@@ -205,6 +210,16 @@ func NewAuthManager(conf AuthManagerConfig) *MenderAuthManager {
 		// regeneration of keys.
 	}
 
+	// Clean up keys that we don't use anymore. This is safe to do even if
+	// rolling back, because the old clients will just fetch a new
+	// AuthToken. Ignore all errors here; this is just cleanup, and it
+	// doesn't matter if it fails.
+	_ = conf.AuthDataStore.WriteTransaction(func(txn store.Transaction) error {
+		_ = txn.Remove(datastore.AuthTokenName)
+		_ = txn.Remove(datastore.AuthTokenCacheInvalidatorName)
+		return nil
+	})
+
 	return mgr
 }
 
@@ -238,41 +253,59 @@ func (m *MenderAuthManager) GetBroadcastMessageChan(name string) <-chan AuthMana
 
 func (m *menderAuthManagerService) registerDBusCallbacks() (unregisterFunc func()) {
 	// GetJwtToken
-	m.dbus.RegisterMethodCallCallback(AuthManagerDBusPath, AuthManagerDBusInterfaceName, "GetJwtToken", func(objectPath, interfaceName, methodName string, parameters string) (interface{}, error) {
-		respChan := make(chan AuthManagerResponse)
-		m.inChan <- AuthManagerRequest{
-			Action:          ActionGetAuthToken,
-			ResponseChannel: respChan,
-		}
-		select {
-		case message := <-respChan:
-			tokenAndServerURL := dbus.TokenAndServerURL{
-				Token:     string(message.AuthToken),
-				ServerURL: m.serverURL,
+	m.dbus.RegisterMethodCallCallback(
+		AuthManagerDBusPath,
+		AuthManagerDBusInterfaceName,
+		"GetJwtToken",
+		func(objectPath, interfaceName, methodName string, parameters string) (interface{}, error) {
+			respChan := make(chan AuthManagerResponse)
+			m.inChan <- AuthManagerRequest{
+				Action:          ActionGetAuthToken,
+				ResponseChannel: respChan,
 			}
-			return tokenAndServerURL, message.Error
-		case <-time.After(5 * time.Second):
-		}
-		return string(noAuthToken), errors.New("timeout when calling GetJwtToken")
-	})
+			select {
+			case message := <-respChan:
+				tokenAndServerURL := dbus.TokenAndServerURL{
+					Token:     string(message.AuthToken),
+					ServerURL: m.localProxy.GetServerUrl(),
+				}
+				return tokenAndServerURL, message.Error
+			case <-time.After(5 * time.Second):
+			}
+			return string(noAuthToken), errors.New("timeout when calling GetJwtToken")
+		},
+	)
 	// FetchJwtToken
-	m.dbus.RegisterMethodCallCallback(AuthManagerDBusPath, AuthManagerDBusInterfaceName, "FetchJwtToken", func(objectPath, interfaceName, methodName string, parameters string) (interface{}, error) {
-		respChan := make(chan AuthManagerResponse)
-		m.inChan <- AuthManagerRequest{
-			Action:          ActionFetchAuthToken,
-			ResponseChannel: respChan,
-		}
-		select {
-		case message := <-respChan:
-			return message.Event == EventFetchAuthToken, message.Error
-		case <-time.After(5 * time.Second):
-		}
-		return false, errors.New("timeout when calling FetchJwtToken")
-	})
+	m.dbus.RegisterMethodCallCallback(
+		AuthManagerDBusPath,
+		AuthManagerDBusInterfaceName,
+		"FetchJwtToken",
+		func(objectPath, interfaceName, methodName string, parameters string) (interface{}, error) {
+			respChan := make(chan AuthManagerResponse)
+			m.inChan <- AuthManagerRequest{
+				Action:          ActionFetchAuthToken,
+				ResponseChannel: respChan,
+			}
+			select {
+			case message := <-respChan:
+				return message.Event == EventFetchAuthToken, message.Error
+			case <-time.After(5 * time.Second):
+			}
+			return false, errors.New("timeout when calling FetchJwtToken")
+		},
+	)
 
 	return func() {
-		m.dbus.UnregisterMethodCallCallback(AuthManagerDBusPath, AuthManagerDBusInterfaceName, "FetchJwtToken")
-		m.dbus.UnregisterMethodCallCallback(AuthManagerDBusPath, AuthManagerDBusInterfaceName, "GetJwtToken")
+		m.dbus.UnregisterMethodCallCallback(
+			AuthManagerDBusPath,
+			AuthManagerDBusInterfaceName,
+			"FetchJwtToken",
+		)
+		m.dbus.UnregisterMethodCallCallback(
+			AuthManagerDBusPath,
+			AuthManagerDBusInterfaceName,
+			"GetJwtToken",
+		)
 	}
 }
 
@@ -283,18 +316,24 @@ func (m *MenderAuthManager) Start() {
 	}
 
 	m.menderAuthManagerService.hasStarted = true
-	go m.menderAuthManagerService.run()
+
+	initDone := make(chan struct{}, 1)
+	go m.menderAuthManagerService.run(initDone)
+
+	// Wait for initialization to finish.
+	<-initDone
+
 	runtime.SetFinalizer(m, func(m *MenderAuthManager) {
 		m.Stop()
 	})
 }
 
 // Run is the main routine of the Mender authorization manager
-func (m *menderAuthManagerService) run() {
+func (m *menderAuthManagerService) run(initDone chan struct{}) {
 	// When we are being stopped, make sure they know that this happened.
 	defer func() {
 		// Checking for panic here is just to avoid deadlocking if we
-		// get an unexpected panic: Let it propogate instead of blocking
+		// get an unexpected panic: Let it propagate instead of blocking
 		// on the channel. If the program is correct, this should never
 		// be non-nil.
 		if recover() == nil {
@@ -303,22 +342,27 @@ func (m *menderAuthManagerService) run() {
 	}()
 
 	// run the DBus interface, if available
-	dbusConn := dbus.Handle(nil)
-	dbusLoop := dbus.MainLoop(nil)
 	if m.dbus != nil {
-		var err error
-		if dbusConn, err = m.dbus.BusGet(dbus.GBusTypeSystem); err == nil {
+		if dbusConn, err := m.dbus.BusGet(dbus.GBusTypeSystem); err == nil {
 			m.dbusConn = dbusConn
 
 			nameGid, err := m.dbus.BusOwnNameOnConnection(dbusConn, AuthManagerDBusObjectName,
 				dbus.DBusNameOwnerFlagsAllowReplacement|dbus.DBusNameOwnerFlagsReplace)
 			if err != nil {
-				log.Errorf("Could not own DBus name '%s': %s", AuthManagerDBusObjectName, err.Error())
+				log.Errorf(
+					"Could not own DBus name '%s': %s",
+					AuthManagerDBusObjectName,
+					err.Error(),
+				)
 				goto mainloop
 			}
 			defer m.dbus.BusUnownName(nameGid)
 
-			intGid, err := m.dbus.BusRegisterInterface(dbusConn, AuthManagerDBusPath, AuthManagerDBusInterface)
+			intGid, err := m.dbus.BusRegisterInterface(
+				dbusConn,
+				AuthManagerDBusPath,
+				AuthManagerDBusInterface,
+			)
 			if err != nil {
 				log.Errorf("Could register DBus interface name '%s' at path '%s': %s",
 					AuthManagerDBusInterface, AuthManagerDBusPath, err.Error())
@@ -329,23 +373,14 @@ func (m *menderAuthManagerService) run() {
 			unregisterFunc := m.registerDBusCallbacks()
 			defer unregisterFunc()
 
-			dbusLoop = m.dbus.MainLoopNew()
+			dbusLoop := m.dbus.MainLoopNew()
 			go m.dbus.MainLoopRun(dbusLoop)
 			defer m.dbus.MainLoopQuit(dbusLoop)
 		}
 	}
 
 mainloop:
-	// Broadcast the TokenStateChange signal once on startup, if we have a
-	// valid token. The reason this is important is that clients that use
-	// the auth DBus API may already have tried calling GetJwtToken
-	// unsuccessfully, and are now waiting for a signal. If we don't
-	// broadcast it on startup, these clients may be left without a token
-	// until it expires and we get a new one, which can take several days.
-	token, err := m.authToken()
-	if err == nil && token != "" {
-		m.broadcastAuthTokenStateChange()
-	}
+	initDone <- struct{}{}
 
 	go m.longRunningWorkerLoop()
 
@@ -375,7 +410,6 @@ mainloop:
 		case <-m.quitReq:
 			running = false
 			m.workerChan <- AuthManagerRequest{}
-			break
 		}
 	}
 }
@@ -383,16 +417,14 @@ mainloop:
 // This is a helper to the main loop, for tasks that may take a long time. It's
 // running in a separate Go routine.
 func (m *menderAuthManagerService) longRunningWorkerLoop() {
-	for {
-		select {
-		case msg := <-m.workerChan:
-			switch msg.Action {
-			case ActionFetchAuthToken:
-				m.fetchAuthToken()
-			case "":
-				// Quit loop.
-				return
-			}
+	for msg := range m.workerChan {
+		switch msg.Action {
+		case ActionFetchAuthToken:
+			m.fetchAuthToken()
+		case "":
+			// Quit loop.
+			return
+
 		}
 	}
 }
@@ -409,16 +441,17 @@ func (m *MenderAuthManager) Stop() {
 	<-m.menderAuthManagerService.quitResp
 	m.menderAuthManagerService.hasStarted = false
 
+	m.localProxy.Stop()
+
 	runtime.SetFinalizer(m, nil)
 }
 
 // getAuthToken returns the cached auth token
 func (m *menderAuthManagerService) getAuthToken(responseChannel chan<- AuthManagerResponse) {
-	authToken, err := m.authToken()
 	msg := AuthManagerResponse{
-		AuthToken: authToken,
+		AuthToken: m.authToken,
+		ServerURL: m.serverURL,
 		Event:     EventGetAuthToken,
-		Error:     err,
 	}
 	responseChannel <- msg
 }
@@ -438,9 +471,9 @@ func (m *menderAuthManagerService) broadcast(message AuthManagerResponse) {
 	if m.dbus != nil {
 		tokenAndServerURL := dbus.TokenAndServerURL{
 			Token:     string(message.AuthToken),
-			ServerURL: m.serverURL,
+			ServerURL: string(message.ServerURL),
 		}
-		m.dbus.EmitSignal(m.dbusConn, "", AuthManagerDBusPath,
+		_ = m.dbus.EmitSignal(m.dbusConn, "", AuthManagerDBusPath,
 			AuthManagerDBusInterfaceName, AuthManagerDBusSignalJwtTokenStateChange,
 			tokenAndServerURL)
 	}
@@ -448,13 +481,30 @@ func (m *menderAuthManagerService) broadcast(message AuthManagerResponse) {
 
 // broadcastAuthTokenStateChange broadcasts the notification to all the subscribers
 func (m *menderAuthManagerService) broadcastAuthTokenStateChange() {
-	authToken, err := m.authToken()
-	if err == nil {
-		m.broadcast(AuthManagerResponse{
-			Event:     EventAuthTokenStateChange,
-			AuthToken: authToken,
-		})
+	m.localProxy.Stop()
+	if m.authToken != "" {
+		// reconfigure proxy
+		err := m.localProxy.Reconfigure(string(m.serverURL), string(m.authToken))
+		if err != nil {
+			log.Errorf(
+				"Could not reconfigure proxy with URL %q and token '%s...'"+
+					" Other applications running on the device won't be able"+
+					" to reach the Mender server. Error: %s",
+				m.serverURL,
+				string(m.authToken)[:7],
+				err.Error(),
+			)
+		} else {
+			m.localProxy.Start()
+
+		}
 	}
+
+	m.broadcast(AuthManagerResponse{
+		Event:     EventAuthTokenStateChange,
+		AuthToken: m.authToken,
+		ServerURL: client.ServerURL(m.localProxy.GetServerUrl()),
+	})
 }
 
 // fetchAuthToken authenticates with the server and retrieve a new auth token, if needed
@@ -465,11 +515,7 @@ func (m *menderAuthManagerService) fetchAuthToken() {
 	resp := AuthManagerResponse{Event: EventFetchAuthToken}
 
 	defer func() {
-		if resp.Error == nil {
-			m.broadcastAuthTokenStateChange()
-		} else {
-			m.broadcast(resp)
-		}
+		m.broadcastAuthTokenStateChange()
 	}()
 
 	if err := m.Bootstrap(); err != nil {
@@ -503,40 +549,36 @@ func (m *menderAuthManagerService) fetchAuthToken() {
 			// SUCCESS!
 			break
 		}
-		prevHost := server.ServerURL
+		log.Errorf("Failed to authorize with %q: %s",
+			server.ServerURL, err.Error())
 		server = serverIterator()
 		if server == nil {
 			break
 		}
-		log.Warnf("Failed to authorize %q; attempting %q.",
-			prevHost, server.ServerURL)
+		log.Infof("Attempting to authorize with %q.",
+			server.ServerURL)
 	}
 	if err != nil {
 		// Generate and report error.
 		errCause := errors.Cause(err)
 		if errCause == client.AuthErrorUnauthorized {
-			// make sure to remove auth token once device is rejected
-			if remErr := m.removeAuthToken(); remErr != nil {
-				log.Warn("can not remove rejected authentication token")
-			}
+			m.authToken = ""
+			m.serverURL = ""
 		}
 		err := NewTransientError(errors.Wrap(err, "authorization request failed"))
 		resp.Error = err
 		return
 	}
 
-	err = m.recvAuthResponse(rsp)
-	if err != nil {
-		err := NewTransientError(errors.Wrap(err, "failed to parse authorization response"))
-		resp.Error = err
+	if len(rsp) == 0 {
+		resp.Error = errors.New("empty auth response data")
 		return
 	}
 
-	// store the current server URL
-	m.serverURL = serverURL
+	m.authToken = client.AuthToken(rsp)
+	m.serverURL = client.ServerURL(serverURL)
 
-	log.Info("successfully received new authorization data")
-	return
+	log.Infof("successfully received new authorization data from server %s", m.serverURL)
 }
 
 // ForceBootstrap forces the bootstrap
@@ -630,44 +672,6 @@ func (m *menderAuthManagerService) MakeAuthRequest() (*client.AuthRequest, error
 	}, nil
 }
 
-// recvAuthResponse processes an auth response
-func (m *menderAuthManagerService) recvAuthResponse(data []byte) error {
-	if len(data) == 0 {
-		return errors.New("empty auth response data")
-	}
-
-	if err := m.store.WriteAll(datastore.AuthTokenName, data); err != nil {
-		return errors.Wrapf(err, "failed to save auth token")
-	}
-	return nil
-}
-
-// AuthToken returns device's authorization token
-func (m *menderAuthManagerService) authToken() (client.AuthToken, error) {
-	data, err := m.store.ReadAll(datastore.AuthTokenName)
-	if err != nil {
-		if os.IsNotExist(err) {
-			return noAuthToken, nil
-		}
-		return noAuthToken, errors.Wrapf(err, "failed to read auth token data")
-	}
-
-	return client.AuthToken(data), nil
-}
-
-// removeAuthToken removes authentication token
-func (m *menderAuthManagerService) removeAuthToken() error {
-	// remove token only if we have one
-	if aToken, err := m.authToken(); err == nil && aToken != noAuthToken {
-		err := m.store.Remove(datastore.AuthTokenName)
-		if err == nil {
-			m.broadcastAuthTokenStateChange()
-		}
-		return err
-	}
-	return nil
-}
-
 // HasKey check if device key is available
 func (m *menderAuthManagerService) HasKey() bool {
 	return m.keyStore.Private() != nil
