@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -24,6 +24,7 @@ import (
 
 	"github.com/pkg/errors"
 
+	"github.com/mendersoftware/mender/app/updatecontrolmap"
 	"github.com/mendersoftware/mender/client"
 	"github.com/mendersoftware/mender/conf"
 	"github.com/mendersoftware/mender/datastore"
@@ -36,9 +37,8 @@ import (
 )
 
 type Controller interface {
-	IsAuthorized() bool
-	Authorize() menderError
-	GetAuthToken() client.AuthToken
+	Authorize() (client.AuthToken, client.ServerURL, error)
+	ClearAuthorization()
 
 	GetControlMapPool() *ControlMapPool
 
@@ -46,9 +46,11 @@ type Controller interface {
 	GetUpdatePollInterval() time.Duration
 	GetInventoryPollInterval() time.Duration
 	GetRetryPollInterval() time.Duration
+	GetRetryPollCount() int
 
 	CheckUpdate() (*datastore.UpdateInfo, menderError)
 	FetchUpdate(url string) (io.ReadCloser, int64, error)
+	RefreshServerUpdateControlMap(deploymentID string) error
 
 	NewStatusReportWrapper(updateId string,
 		stateId datastore.MenderState) *client.StatusReportWrapper
@@ -68,7 +70,8 @@ type Controller interface {
 }
 
 var (
-	errNoArtifactName = errors.New("cannot determine current artifact name")
+	errNoArtifactName        = errors.New("cannot determine current artifact name")
+	errControlMapIDMismatchF = "Mismatched control map ID: %s and deployment ID: %s"
 )
 
 var (
@@ -116,8 +119,12 @@ type Mender struct {
 	state               State
 	stateScriptExecutor statescript.Executor
 	authManager         AuthManager
-	api                 *client.ApiClient
-	controlMapPool      *ControlMapPool
+	// Used for requesting API URLs.
+	api client.AuthorizedApiRequester
+	// Used for downloading artifacts.
+	download client.ApiRequester
+
+	controlMapPool *ControlMapPool
 }
 
 type MenderPieces struct {
@@ -127,11 +134,6 @@ type MenderPieces struct {
 }
 
 func NewMender(config *conf.MenderConfig, pieces MenderPieces) (*Mender, error) {
-	api, err := client.New(config.GetHttpConfig())
-	if err != nil {
-		return nil, errors.Wrap(err, "error creating HTTP client")
-	}
-
 	stateScrExec := dev.NewStateScriptExecutor(config)
 
 	controlMapPool := NewControlMap(
@@ -146,52 +148,31 @@ func NewMender(config *conf.MenderConfig, pieces MenderPieces) (*Mender, error)
 		state:               States.Init,
 		stateScriptExecutor: stateScrExec,
 		authManager:         pieces.AuthManager,
-		api:                 api,
 		controlMapPool:      controlMapPool,
 	}
 
+	api, err := client.NewReauthorizingClient(config.GetHttpConfig(), m.Authorize)
+	if err != nil {
+		return nil, errors.Wrap(err, "error creating HTTP API client")
+	}
+	m.api = api
+
+	m.download, err = client.NewApiClient(config.GetHttpConfig())
+	if err != nil {
+		return nil, errors.Wrap(err, "error creating HTTP download client")
+	}
+
 	return m, nil
 }
 
-// cache authorization code
-func (m *Mender) loadAuth() (client.AuthToken, menderError) {
-	inChan := m.authManager.GetInMessageChan()
-	respChan := make(chan AuthManagerResponse)
-
-	// request
-	inChan <- AuthManagerRequest{
-		Action:          ActionGetAuthToken,
-		ResponseChannel: respChan,
-	}
-
-	// response
-	resp := <-respChan
-	if resp.Error != nil {
-		return noAuthToken, NewTransientError(resp.Error)
-	}
-
-	return resp.AuthToken, nil
-}
-
-func (m *Mender) IsAuthorized() bool {
-	authToken, err := m.loadAuth()
-	if err != nil {
-		return false
-	}
-	if authToken != noAuthToken {
-		return true
-	}
-	return false
-}
-
-func (m *Mender) Authorize() menderError {
+func (m *Mender) Authorize() (client.AuthToken, client.ServerURL, error) {
 	inChan := m.authManager.GetInMessageChan()
 	broadcastChan := m.authManager.GetBroadcastMessageChan(authManagerChannelName)
 	respChan := make(chan AuthManagerResponse)
 
 	// drain the broadcast channel
 	select {
-	case _ = <-broadcastChan:
+	case <-broadcastChan:
 	default:
 	}
 
@@ -204,24 +185,30 @@ func (m *Mender) Authorize() menderError {
 	// response
 	resp := <-respChan
 	if resp.Error != nil {
-		return NewTransientError(resp.Error)
+		return "", "", NewTransientError(resp.Error)
 	}
 
 	// wait for the broadcast notification
 	resp, ok := <-broadcastChan
-	if !ok || resp.Error != nil {
-		return NewTransientError(errors.Wrap(resp.Error, "authorization request failed"))
+	if !ok {
+		return "", "", NewTransientError(
+			errors.New("authorization request failed: channel read failed"),
+		)
+	} else if resp.Error != nil {
+		return "", "", NewTransientError(
+			errors.Wrap(resp.Error, "authorization request failed"),
+		)
+	} else if len(resp.AuthToken) == 0 || len(resp.ServerURL) == 0 {
+		return "", "", NewTransientError(
+			errors.New("authorization request failed"),
+		)
 	}
 
-	return nil
+	return resp.AuthToken, resp.ServerURL, nil
 }
 
-func (m *Mender) GetAuthToken() client.AuthToken {
-	authToken, err := m.loadAuth()
-	if err != nil {
-		log.Errorf("Could not load auth token: %s", err.Error())
-	}
-	return authToken
+func (m *Mender) ClearAuthorization() {
+	m.api.ClearAuthorization()
 }
 
 func (m *Mender) GetControlMapPool() *ControlMapPool {
@@ -229,7 +216,7 @@ func (m *Mender) GetControlMapPool() *ControlMapPool {
 }
 
 func (m *Mender) FetchUpdate(url string) (io.ReadCloser, int64, error) {
-	return m.updater.FetchUpdate(m.api, url, m.GetRetryPollInterval())
+	return m.updater.FetchUpdate(m.download, url, m.GetRetryPollInterval())
 }
 
 func verifyArtifactDependencies(
@@ -243,7 +230,7 @@ func verifyArtifactDependencies(
 			continue
 		}
 		if p, ok := provides[key]; ok {
-			switch depend.(type) {
+			switch pVal := depend.(type) {
 			case []interface{}:
 				if ok, err := utils.ElemInSlice(depend, p); ok {
 					continue
@@ -260,7 +247,7 @@ func verifyArtifactDependencies(
 					continue
 				}
 			case string:
-				if p == depend.(string) {
+				if p == pVal {
 					continue
 				}
 			default:
@@ -289,7 +276,15 @@ func (m *Mender) CheckUpdate() (*datastore.UpdateInfo, menderError) {
 		if err == nil {
 			err = errors.New("artifact name is empty")
 		}
-		return nil, NewTransientError(fmt.Errorf("could not read the Artifact name. This is a necessary condition in order for a Mender update to finish safely. Please give the current Artifact a name (This can be done by adding a name to the file /etc/mender/artifact_info) err: %v", err))
+		return nil, NewTransientError(
+			fmt.Errorf(
+				"could not read the Artifact name. This is a necessary condition in order for"+
+					" a Mender update to finish safely. Please give the current Artifact a name"+
+					" (This can be done by adding a name to the file /etc/mender/artifact_info)"+
+					" err: %v",
+				err,
+			),
+		)
 	}
 
 	deviceType, err := m.GetDeviceType()
@@ -299,13 +294,14 @@ func (m *Mender) CheckUpdate() (*datastore.UpdateInfo, menderError) {
 	}
 	provides, err := m.DeviceManager.GetProvides()
 	if err != nil {
-		log.Errorf("Failed to load the device provides parameters from the datastore. Error: %v. Continuing...",
-			err)
+		log.Errorf(
+			"Failed to load the device provides parameters from the datastore. Error: %v."+
+				" Continuing...",
+			err,
+		)
 	}
 	haveUpdate, err := m.updater.GetScheduledUpdate(
-		m.api.Request(m.GetAuthToken(),
-			nextServerIterator(m.Config),
-			reauthorize(m)),
+		m.api,
 		m.Config.Servers[0].ServerURL,
 		&client.CurrentUpdate{
 			Artifact:   currentArtifactName,
@@ -316,13 +312,6 @@ func (m *Mender) CheckUpdate() (*datastore.UpdateInfo, menderError) {
 	ur, urOk := haveUpdate.(client.UpdateResponse)
 
 	if err != nil {
-		// remove authentication token if device is not authorized
-		errCause := errors.Cause(err)
-		if errCause == client.ErrNotAuthorized {
-			if authErr := m.Authorize(); authErr != nil {
-				log.Warn("can not perform reauthorization")
-			}
-		}
 		if errors.Is(err, client.ErrNoDeploymentAvailable) {
 			return ur.UpdateInfo, NewTransientError(err)
 		}
@@ -337,42 +326,64 @@ func (m *Mender) CheckUpdate() (*datastore.UpdateInfo, menderError) {
 
 	if !urOk {
 		err = fmt.Errorf(
-			"The update data received is unexpectedly the wrong type %T. Expected 'client.UpdateResponse'",
-			haveUpdate)
+			"The update data received is unexpectedly the wrong type %T. Expected"+
+				" 'client.UpdateResponse'",
+			haveUpdate,
+		)
 		return nil, NewTransientError(err)
 	}
 
 	log.Debugf("Received update response: %v", ur)
-	if err = m.handleControlMap(&ur); err != nil {
+	if err = m.HandleControlMap(ur.ID, ur.UpdateControlMap); err != nil {
 		return ur.UpdateInfo, NewTransientError(err)
 	}
 
 	if ur.UpdateInfo.ArtifactName() == currentArtifactName {
-		log.Info("Attempting to upgrade to currently installed artifact name, not performing upgrade.")
+		log.Info(
+			"Attempting to upgrade to currently installed artifact name, not performing upgrade.",
+		)
 		return ur.UpdateInfo, NewTransientError(os.ErrExist)
 	}
 
 	return ur.UpdateInfo, nil
 }
 
-func (m *Mender) handleControlMap(data *client.UpdateResponse) error {
-	if data.UpdateControlMap != nil {
-		if data.UpdateControlMap.ID != "" {
-			if data.UpdateControlMap.ID != data.UpdateInfo.ID {
+// RefreshServerUpdateControlMap updates the control maps from the server during
+// a deployment.
+func (m *Mender) RefreshServerUpdateControlMap(deploymentID string) error {
+	cm, err := client.GetUpdateControlMap(
+		m.api,
+		m.Config.Servers[0].ServerURL,
+		deploymentID,
+	)
+	if err != nil {
+		return err
+	}
+
+	return m.HandleControlMap(deploymentID, cm)
+}
+
+func (m *Mender) HandleControlMap(
+	deploymentID string,
+	updateControlMap *updatecontrolmap.UpdateControlMap,
+) error {
+	if updateControlMap != nil {
+		if updateControlMap.ID != "" {
+			if updateControlMap.ID != deploymentID {
 				return NewTransientError(
-					fmt.Errorf("Mismatched control map ID: %s and deployment ID: %s",
-						data.UpdateControlMap.ID, data.UpdateInfo.ID))
+					fmt.Errorf(errControlMapIDMismatchF,
+						updateControlMap.ID, deploymentID))
 			}
 		} else {
-			data.UpdateControlMap.ID = data.UpdateInfo.ID
+			updateControlMap.ID = deploymentID
 		}
 		m.controlMapPool.InsertReplaceAllPriorities(
-			data.UpdateControlMap.Stamp(
+			updateControlMap.Stamp(
 				m.DeviceManager.Config.
 					MenderConfigFromFile.
 					GetUpdateControlMapExpirationTimeSeconds()))
 	} else {
-		m.controlMapPool.DeleteAllPriorities(data.UpdateInfo.ID)
+		m.controlMapPool.DeleteAllPriorities(deploymentID)
 	}
 	return nil
 }
@@ -381,7 +392,7 @@ func (m *Mender) NewStatusReportWrapper(updateId string,
 	stateId datastore.MenderState) *client.StatusReportWrapper {
 
 	return &client.StatusReportWrapper{
-		API: m.api.Request(m.GetAuthToken(), nextServerIterator(m.Config), reauthorize(m)),
+		API: m.api,
 		URL: m.Config.Servers[0].ServerURL,
 		Report: client.StatusReport{
 			DeploymentID: updateId,
@@ -392,20 +403,18 @@ func (m *Mender) NewStatusReportWrapper(updateId string,
 
 func (m *Mender) ReportUpdateStatus(update *datastore.UpdateInfo, status string) menderError {
 	s := client.NewStatus()
-	err := s.Report(m.api.Request(m.GetAuthToken(), nextServerIterator(m.Config), reauthorize(m)), m.Config.Servers[0].ServerURL,
+	err := s.Report(
+		m.api,
+		m.Config.Servers[0].ServerURL,
 		client.StatusReport{
 			DeploymentID: update.ID,
 			Status:       status,
-		})
+		},
+	)
 	if err != nil {
 		log.Error("error reporting update status: ", err)
-		// remove authentication token if device is not authorized
 		errCause := errors.Cause(err)
-		if errCause == client.ErrNotAuthorized {
-			if authErr := m.Authorize(); authErr != nil {
-				log.Warn("can not perform reauthorization")
-			}
-		} else if errCause == client.ErrDeploymentAborted {
+		if errCause == client.ErrDeploymentAborted {
 			return NewFatalError(err)
 		}
 		return NewTransientError(err)
@@ -413,28 +422,16 @@ func (m *Mender) ReportUpdateStatus(update *datastore.UpdateInfo, status string)
 	return nil
 }
 
-// see client.go: ApiRequest.Do()
-
-// reauthorize is a closure very similar to mender.Authorize(), but instead of
-// walking through all servers in the conf.MenderConfig.Servers list, it only tries
-// serverURL.
-func reauthorize(m *Mender) func(string) (client.AuthToken, error) {
-	// force reauthorization
-	return func(serverURL string) (client.AuthToken, error) {
-		if err := m.Authorize(); err != nil {
-			return noAuthToken, err
-		}
-		return m.loadAuth()
-	}
-}
-
 func (m *Mender) UploadLog(update *datastore.UpdateInfo, logs []byte) menderError {
 	s := client.NewLog()
-	err := s.Upload(m.api.Request(m.GetAuthToken(), nextServerIterator(m.Config), reauthorize(m)), m.Config.Servers[0].ServerURL,
+	err := s.Upload(
+		m.api,
+		m.Config.Servers[0].ServerURL,
 		client.LogData{
 			DeploymentID: update.ID,
 			Messages:     logs,
-		})
+		},
+	)
 	if err != nil {
 		log.Error("error uploading logs: ", err)
 		return NewTransientError(err)
@@ -469,6 +466,10 @@ func (m *Mender) GetRetryPollInterval() time.Duration {
 	return t
 }
 
+func (m *Mender) GetRetryPollCount() int {
+	return m.Config.RetryPollCount
+}
+
 func (m *Mender) SetNextState(s State) {
 	m.state = s
 }
@@ -535,7 +536,8 @@ func transitionState(to State, ctx *StateContext, c Controller) (State, bool) {
 			_ = from.Transition().Error(c.GetScriptExecutor(), report)
 		} else {
 			// do transition to ordinary state
-			if err := from.Transition().Leave(c.GetScriptExecutor(), report, ctx.Store); err != nil {
+			if err := from.Transition().
+				Leave(c.GetScriptExecutor(), report, ctx.Store); err != nil {
 				merr := NewTransientError(fmt.Errorf(
 					"error executing leave script for %s state: %s",
 					from.Id(), err.Error()))
@@ -596,7 +598,13 @@ func (m *Mender) InventoryRefresh() error {
 		if err == nil {
 			err = errors.New("Artifact name is empty")
 		}
-		errstr := fmt.Sprintf("could not read the artifact name. This is a necessary condition in order for a Mender update to finish safely. Please give the current Artifact a name (This can be done by adding a name to the file /etc/mender/artifact_info) err: %v", err)
+		errstr := fmt.Sprintf(
+			"could not read the artifact name. This is a necessary condition in order for"+
+				" a Mender update to finish safely. Please give the current Artifact a name"+
+				" (This can be done by adding a name to the file /etc/mender/artifact_info)"+
+				" err: %v",
+			err,
+		)
 		return errors.Wrap(errNoArtifactName, errstr)
 	}
 
@@ -608,7 +616,11 @@ func (m *Mender) InventoryRefresh() error {
 
 	deviceType, err := m.GetDeviceType()
 	if err != nil {
-		log.Errorf("Unable to verify the existing hardware. Update will continue anyways: %v : %v", m.Config.DeviceTypeFile, err)
+		log.Errorf(
+			"Unable to verify the existing hardware. Update will continue anyways: %v : %v",
+			m.Config.DeviceTypeFile,
+			err,
+		)
 	}
 	reqAttr := []client.InventoryAttribute{
 		{Name: "device_type", Value: deviceType},
@@ -626,7 +638,7 @@ func (m *Mender) InventoryRefresh() error {
 		return nil
 	}
 
-	err = ic.Submit(m.api.Request(m.GetAuthToken(), nextServerIterator(m.Config), reauthorize(m)), m.Config.Servers[0].ServerURL, idata)
+	err = ic.Submit(m.api, m.Config.Servers[0].ServerURL, idata)
 	if err != nil {
 		return errors.Wrapf(err, "failed to submit inventory data")
 	}
@@ -637,3 +649,24 @@ func (m *Mender) InventoryRefresh() error {
 func (m *Mender) CheckScriptsCompatibility() error {
 	return m.stateScriptExecutor.CheckRootfsScriptsVersion()
 }
+
+func verifyAndSetArtifactNameInProvides(
+	provides map[string]string,
+	getArtifactName func() (string, error),
+) (map[string]string, error) {
+	if _, ok := provides["artifact_name"]; !ok {
+		artifactName, err := getArtifactName()
+		if err != nil || artifactName == "" {
+			log.Error("could not get the current Artifact name")
+			if err == nil {
+				err = errors.New("artifact name is empty")
+			}
+			return provides, err
+		}
+		if provides == nil {
+			provides = make(map[string]string)
+		}
+		provides["artifact_name"] = artifactName
+	}
+	return provides, nil
+}
