@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -321,20 +321,32 @@ func TestDoManualUpdateArtifactV3Dependencies(t *testing.T) {
 		ArtifactScriptsPath: tmpdir,
 	}
 
+	testDevMgr := getTestDeviceManager(fakeDev, &config, deviceType, dbdir)
+
 	// First check that unmet dependencies returns an error, and add
 	// one dependency at the time untill the artifact should be accepted.
-	testDevMgr := getTestDeviceManager(fakeDev, &config, deviceType, dbdir)
 	err = DoStandaloneInstall(testDevMgr,
 		imageFileName, client.Config{},
 		nil, dev.NewStateScriptExecutor(&config), false)
 	assert.Error(t, err)
 
+	// Try with existing, but null typeInfoProvides.
+	testDevMgr.Store.WriteAll(
+		datastore.ArtifactTypeInfoProvidesKey, []byte("null"))
+	err = DoStandaloneInstall(testDevMgr,
+		imageFileName, client.Config{},
+		nil, dev.NewStateScriptExecutor(&config), false)
+	assert.Error(t, err)
+
+	// Try with artifact_name inserted.
 	testDevMgr.Store.WriteAll(datastore.ArtifactNameKey,
 		[]byte("OldArtifact"))
 	err = DoStandaloneInstall(testDevMgr,
 		imageFileName, client.Config{},
 		nil, dev.NewStateScriptExecutor(&config), false)
 	assert.Error(t, err)
+
+	// Try with artifact_group inserted.
 	testDevMgr.Store.WriteAll(datastore.ArtifactGroupKey,
 		[]byte("testGroup"))
 	err = DoStandaloneInstall(testDevMgr,
@@ -342,6 +354,7 @@ func TestDoManualUpdateArtifactV3Dependencies(t *testing.T) {
 		nil, dev.NewStateScriptExecutor(&config), false)
 	assert.Error(t, err)
 
+	// Try with typeInfoProvides inserted.
 	typeProvidesBuf, err := json.Marshal(typeInfoDepends)
 	assert.NoError(t, err)
 	testDevMgr.Store.WriteAll(
@@ -1105,7 +1118,12 @@ func TestStandaloneStoreAndRestore(t *testing.T) {
 	for _, test := range tests {
 		t.Logf("Running test: %v", test)
 		// First transform
-		assert.NoError(t, storeStandaloneData(dbstore, test.sd), "Failed to store the standaloneData: %v", test.sd)
+		assert.NoError(
+			t,
+			storeStandaloneData(dbstore, test.sd),
+			"Failed to store the standaloneData: %v",
+			test.sd,
+		)
 
 		sd, err := restoreStandaloneData(tmgr)
 		assert.NoError(t, err, "Failed to restore the standaloneData")
@@ -1376,7 +1394,12 @@ func TestStandaloneInstallProvides(t *testing.T) {
 
 			artPath := path.Join(tmpdir, "artifact.mender")
 
-			device, stateExec := standaloneInstallSetup(t, tmpdir, &tests.TestModuleAttr{}, c.overrides)
+			device, stateExec := standaloneInstallSetup(
+				t,
+				tmpdir,
+				&tests.TestModuleAttr{},
+				c.overrides,
+			)
 
 			artifactGroup, hasArtifactGroup := c.preexistingProvides["artifact_group"]
 			if hasArtifactGroup {
