@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -17,6 +17,7 @@ import (
 	"fmt"
 	"io"
 	"os"
+	"sync"
 	"time"
 
 	"github.com/pkg/errors"
@@ -34,6 +35,49 @@ const (
 		"by currently installed artifact (%v != %v)."
 )
 
+type timerPool sync.Pool
+
+func (t *timerPool) Get(wait time.Duration) *time.Timer {
+	timer := (*sync.Pool)(t).Get().(*time.Timer)
+	// Make sure timer is properly reset before use
+	if !timer.Stop() {
+		select {
+		case <-timer.C:
+		default:
+		}
+	}
+	timer.Reset(wait)
+	return timer
+}
+
+func (t *timerPool) Put(timer *time.Timer) {
+	if !timer.Stop() {
+		select {
+		case <-timer.C:
+		default:
+		}
+	}
+	(*sync.Pool)(t).Put(timer)
+}
+
+func newTimerPool() *timerPool {
+	return &timerPool{
+		// New returns a new stopped timer (the timer must be reset before use)
+		New: func() interface{} {
+			t := time.NewTimer(time.Hour)
+			if !t.Stop() {
+				select {
+				case <-t.C:
+				default:
+				}
+			}
+			return t
+		},
+	}
+}
+
+var timers = newTimerPool()
+
 // StateContext carrying over data that may be used by all state handlers
 type StateContext struct {
 	// data store access
@@ -42,8 +86,10 @@ type StateContext struct {
 	WakeupChan                 chan bool
 	lastUpdateCheckAttempt     time.Time
 	lastInventoryUpdateAttempt time.Time
-	lastAuthorizeAttempt       time.Time
 	fetchInstallAttempts       int
+	controlMapFetchAttempts    int
+	inventoryUpdateAttempts    int
+	nextAttemptAt              time.Time
 	pauseReported              map[string]bool
 }
 
@@ -57,8 +103,6 @@ type StateRunner interface {
 }
 
 type StateCollection struct {
-	Authorize       *authorizeState
-	AuthorizeWait   *authorizeWaitState
 	CheckWait       *checkWaitState
 	Final           *finalState
 	Idle            *idleState
@@ -69,14 +113,7 @@ type StateCollection struct {
 
 // Exposed state variables.
 var States = StateCollection{
-	Authorize: &authorizeState{
-		baseState{
-			id: datastore.MenderStateAuthorize,
-			t:  ToSync,
-		},
-	},
-	AuthorizeWait: NewAuthorizeWaitState().(*authorizeWaitState),
-	CheckWait:     NewCheckWaitState().(*checkWaitState),
+	CheckWait: NewCheckWaitState().(*checkWaitState),
 	Final: &finalState{
 		baseState{
 			id: datastore.MenderStateDone,
@@ -95,12 +132,7 @@ var States = StateCollection{
 			t:  ToIdle,
 		},
 	},
-	InventoryUpdate: &inventoryUpdateState{
-		baseState{
-			id: datastore.MenderStateInventoryUpdate,
-			t:  ToSync,
-		},
-	},
+	InventoryUpdate: NewInventoryUpdateState().(*inventoryUpdateState),
 	UpdateCheck: &updateCheckState{
 		baseState{
 			id: datastore.MenderStateUpdateCheck,
@@ -189,19 +221,19 @@ func NewWaitState(id datastore.MenderState, t Transition) *waitState {
 // has completed. If wait was interrupted returns (`same`, true)
 func (ws *waitState) Wait(next, same State,
 	wait time.Duration, wakeup chan bool) (State, bool) {
-	ticker := time.NewTicker(wait)
+	timer := timers.Get(wait)
 	ws.wakeup = wakeup
 
-	defer ticker.Stop()
+	defer timers.Put(timer)
 	select {
-	case <-ticker.C:
-		log.Debugf("Wait complete")
+	case <-timer.C:
+		log.Debug("Wait complete")
 		return next, false
 	case <-ws.wakeup:
 		log.Info("Forced wake-up from sleep")
 		return next, false
 	case <-ws.cancel:
-		log.Infof("Wait canceled")
+		log.Info("Wait canceled")
 	}
 	return same, true
 }
@@ -256,19 +288,15 @@ type idleState struct {
 
 func (i *idleState) Handle(ctx *StateContext, c Controller) (State, bool) {
 	// stop deployment logging
-	DeploymentLogger.Disable()
+	_ = DeploymentLogger.Disable()
 
 	// cleanup state-data if any data is still present after an update
-	RemoveStateData(ctx.Store)
+	_ = RemoveStateData(ctx.Store)
 
 	// Remove the expired UpdateControlMaps from the expired pool
 	c.GetControlMapPool().ClearExpired()
 
-	// check if client is authorized
-	if c.IsAuthorized() {
-		return States.CheckWait, false
-	}
-	return States.AuthorizeWait, false
+	return States.CheckWait, false
 }
 
 type initState struct {
@@ -412,69 +440,6 @@ func (i *initState) getNextState(ctx *StateContext, sd *datastore.StateData,
 	}
 }
 
-type authorizeWaitState struct {
-	baseState
-	WaitState
-}
-
-func NewAuthorizeWaitState() State {
-	return &authorizeWaitState{
-		baseState: baseState{
-			id: datastore.MenderStateAuthorizeWait,
-			t:  ToIdle,
-		},
-		WaitState: NewWaitState(datastore.MenderStateAuthorizeWait, ToIdle),
-	}
-}
-
-func (a *authorizeWaitState) Cancel() bool {
-	return a.WaitState.Cancel()
-}
-
-func (a *authorizeWaitState) Handle(ctx *StateContext, c Controller) (State, bool) {
-	log.Debugf("Handle authorize wait state")
-
-	attempt := ctx.lastAuthorizeAttempt.Add(c.GetRetryPollInterval())
-
-	now := time.Now()
-	var wait time.Duration
-	if attempt.After(now) {
-		wait = attempt.Sub(now)
-	}
-
-	log.Debugf("Wait %v before next authorization attempt", wait)
-
-	if wait == 0 {
-		ctx.lastAuthorizeAttempt = now
-		return States.Authorize, false
-	}
-
-	ctx.lastAuthorizeAttempt = attempt
-	return a.Wait(States.Authorize, a, wait, ctx.WakeupChan)
-}
-
-type authorizeState struct {
-	baseState
-}
-
-func (a *authorizeState) Handle(ctx *StateContext, c Controller) (State, bool) {
-	// stop deployment logging
-	DeploymentLogger.Disable()
-
-	log.Debugf("Handle authorize state")
-	if err := c.Authorize(); err != nil {
-		log.Errorf("Authorize failed: %v", err)
-		if !err.IsFatal() {
-			return States.AuthorizeWait, false
-		}
-		return NewErrorState(err), false
-	}
-	// if everything is OK we should let Mender figure out what to do
-	// in MenderStateCheckWait state
-	log.Info("Server authorization successful")
-	return States.CheckWait, false
-}
-
 type updateCommitState struct {
 	*updateState
 	reportTries int
@@ -503,6 +468,11 @@ func (uc *updateCommitState) Handle(ctx *StateContext, c Controller) (State, boo
 		return uc.HandleError(ctx, c, merr)
 	}
 
+	// Clear our authorization in order to force reauthorization, since the
+	// update may have changed the conditions for authorization, such as
+	// keys or the identity script, and this will check that it works.
+	c.ClearAuthorization()
+
 	// A last status report to the server before committing. This is most
 	// likely a repeat of the previous status, but the real motivation
 	// behind it is to find out whether the server cancelled the deployment
@@ -583,7 +553,10 @@ func NewUpdatePreCommitStatusReportRetryState(returnToState State, reportTries i
 	}
 }
 
-func (usr *updatePreCommitStatusReportRetryState) Handle(ctx *StateContext, c Controller) (State, bool) {
+func (usr *updatePreCommitStatusReportRetryState) Handle(
+	ctx *StateContext,
+	c Controller,
+) (State, bool) {
 	maxTrySending :=
 		maxSendingAttempts(c.GetUpdatePollInterval(),
 			c.GetRetryPollInterval(), minReportSendRetries)
@@ -635,7 +608,11 @@ func (uc *updateAfterFirstCommitState) Handle(ctx *StateContext, c Controller) (
 	return NewUpdateAfterCommitState(uc.Update()), false
 }
 
-func (uc *updateAfterFirstCommitState) HandleError(ctx *StateContext, c Controller, merr menderError) (State, bool) {
+func (uc *updateAfterFirstCommitState) HandleError(
+	ctx *StateContext,
+	c Controller,
+	merr menderError,
+) (State, bool) {
 	log.Error(merr.Error())
 
 	// Too late to back out now. Just report the error, but do not try to roll back.
@@ -662,7 +639,11 @@ func (uc *updateAfterCommitState) Handle(ctx *StateContext, c Controller) (State
 	return NewUpdateCleanupState(uc.Update(), client.StatusSuccess), false
 }
 
-func (uc *updateAfterCommitState) HandleError(ctx *StateContext, c Controller, merr menderError) (State, bool) {
+func (uc *updateAfterCommitState) HandleError(
+	ctx *StateContext,
+	c Controller,
+	merr menderError,
+) (State, bool) {
 	log.Error(merr.Error())
 
 	// Too late to back out now. Just report the error, but do not try to roll back.
@@ -697,7 +678,7 @@ func (u *updateCheckState) Handle(ctx *StateContext, c Controller) (State, bool)
 			// itself, not the network. Fail the update immediately,
 			// because this is not something we expect to recover
 			// from.
-			DeploymentLogger.Enable(update.ID)
+			_ = DeploymentLogger.Enable(update.ID)
 			log.Error(err.Error())
 			return NewUpdateStatusReportState(update, client.StatusFailure), false
 		}
@@ -819,7 +800,7 @@ func (u *updateStoreState) Handle(ctx *StateContext, c Controller) (State, bool)
 	installer, err := c.ReadArtifactHeaders(u.imagein)
 	if err != nil {
 		log.Errorf("Fetching Artifact headers failed: %s", err)
-		return NewUpdateCleanupState(&u.update, client.StatusFailure), false
+		return NewUpdateStatusReportState(&u.update, client.StatusFailure), false
 	}
 
 	installers := c.GetInstallers()
@@ -835,7 +816,7 @@ func (u *updateStoreState) Handle(ctx *StateContext, c Controller) (State, bool)
 			client.StatusFailure), false
 	}
 
-	err = u.maybeVerifyArtifactDependsAndProvides(ctx, installer)
+	err = u.maybeVerifyArtifactDependsAndProvides(ctx, installer, c)
 	if err != nil {
 		return NewUpdateStatusReportState(u.Update(),
 			client.StatusFailure), false
@@ -880,7 +861,7 @@ func (u *updateStoreState) Handle(ctx *StateContext, c Controller) (State, bool)
 }
 
 func (u *updateStoreState) maybeVerifyArtifactDependsAndProvides(
-	ctx *StateContext, installer *installer.Installer) error {
+	ctx *StateContext, installer *installer.Installer, c Controller) error {
 	// For artifact version >= 3 we need to fetch the artifact provides of
 	// the previous artifact from the datastore, and verify that the
 	// provides match artifact dependencies.
@@ -895,6 +876,13 @@ func (u *updateStoreState) maybeVerifyArtifactDependsAndProvides(
 			log.Error(err.Error())
 			return err
 		}
+		if provides, err = verifyAndSetArtifactNameInProvides(
+			provides,
+			c.GetCurrentArtifactName,
+		); err != nil {
+			log.Error(err.Error())
+			return err
+		}
 		if err = verifyArtifactDependencies(
 			depends, provides); err != nil {
 			log.Error(err.Error())
@@ -927,7 +915,10 @@ func (u *updateStoreState) maybeVerifyArtifactDependsAndProvides(
 	return nil
 }
 
-func (u *updateStoreState) handleSupportsRollback(ctx *StateContext, c Controller) (bool, State, bool) {
+func (u *updateStoreState) handleSupportsRollback(
+	ctx *StateContext,
+	c Controller,
+) (bool, State, bool) {
 	for _, i := range c.GetInstallers() {
 		supportsRollback, err := i.SupportsRollback()
 		if err != nil {
@@ -954,15 +945,25 @@ func (u *updateStoreState) handleSupportsRollback(ctx *StateContext, c Controlle
 	}, true)
 	if err != nil {
 		log.Error("Could not write state data to persistent storage: ", err.Error())
-		state, cancelled := handleStateDataError(ctx, NewUpdateErrorState(NewTransientError(err), &u.update),
-			false, u.Id(), &u.update, err)
+		state, cancelled := handleStateDataError(
+			ctx,
+			NewUpdateErrorState(NewTransientError(err), &u.update),
+			false,
+			u.Id(),
+			&u.update,
+			err,
+		)
 		return false, state, cancelled
 	}
 
 	return true, nil, false
 }
 
-func (is *updateStoreState) HandleError(ctx *StateContext, c Controller, merr menderError) (State, bool) {
+func (is *updateStoreState) HandleError(
+	ctx *StateContext,
+	c Controller,
+	merr menderError,
+) (State, bool) {
 	log.Error(merr.Error())
 	return NewUpdateCleanupState(is.Update(), client.StatusFailure), false
 }
@@ -980,10 +981,14 @@ func NewUpdateAfterStoreState(update *datastore.UpdateInfo) State {
 
 func (s *updateAfterStoreState) Handle(ctx *StateContext, c Controller) (State, bool) {
 	// This state only exists to run Download_Leave.
-	return NewControlMapState(NewUpdateInstallState(s.Update())), false
+	return NewFetchControlMapState(NewUpdateInstallState(s.Update()), nil), false
 }
 
-func (s *updateAfterStoreState) HandleError(ctx *StateContext, c Controller, merr menderError) (State, bool) {
+func (s *updateAfterStoreState) HandleError(
+	ctx *StateContext,
+	c Controller,
+	merr menderError,
+) (State, bool) {
 	log.Error(merr.Error())
 	return NewUpdateCleanupState(s.Update(), client.StatusFailure), false
 }
@@ -1036,7 +1041,7 @@ func (is *updateInstallState) Handle(ctx *StateContext, c Controller) (State, bo
 
 		case datastore.RebootTypeCustom, datastore.RebootTypeAutomatic:
 			// Go to reboot state if at least one payload requested it.
-			return NewControlMapState(NewUpdateRebootState(is.Update())), false
+			return is.createRebootState(), false
 
 		default:
 			return is.HandleError(ctx, c, NewTransientError(errors.New(
@@ -1045,10 +1050,18 @@ func (is *updateInstallState) Handle(ctx *StateContext, c Controller) (State, bo
 	}
 
 	// No reboot requests, go to commit state.
-	return NewControlMapState(NewUpdateCommitState(is.Update())), false
+	return NewFetchControlMapState(NewUpdateCommitState(is.Update()), nil), false
 }
 
-func (is *updateInstallState) handleRebootType(ctx *StateContext, c Controller) (bool, State, bool) {
+func (is *updateInstallState) createRebootState() State {
+	return NewFetchControlMapState(NewUpdateRebootState(is.Update()),
+		NewUpdateRebootPauseRequestedState(is.Update()))
+}
+
+func (is *updateInstallState) handleRebootType(
+	ctx *StateContext,
+	c Controller,
+) (bool, State, bool) {
 	for n, i := range c.GetInstallers() {
 		needsReboot, err := i.NeedsReboot()
 		if err != nil {
@@ -1117,7 +1130,11 @@ func (fir *fetchStoreRetryState) Cancel() bool {
 func (fir *fetchStoreRetryState) Handle(ctx *StateContext, c Controller) (State, bool) {
 	log.Debugf("Handle fetch install retry state")
 
-	intvl, err := client.GetExponentialBackoffTime(ctx.fetchInstallAttempts, c.GetUpdatePollInterval())
+	intvl, err := client.GetExponentialBackoffTime(
+		ctx.fetchInstallAttempts,
+		c.GetUpdatePollInterval(),
+		c.GetRetryPollCount(),
+	)
 	if err != nil {
 		if fir.err != nil {
 			return NewUpdateErrorState(
@@ -1134,6 +1151,74 @@ func (fir *fetchStoreRetryState) Handle(ctx *StateContext, c Controller) (State,
 	return fir.Wait(NewUpdateFetchState(&fir.update), fir, intvl, ctx.WakeupChan)
 }
 
+type inventoryUpdateRetry struct {
+	baseState
+	WaitState
+	from State
+	err  error
+}
+
+func NewInventoryUpdateRetryState(from State,
+	err error) State {
+	return &inventoryUpdateRetry{
+		baseState: baseState{
+			id: datastore.MenderStateInventoryUpdateRetryWait,
+			t:  ToSync,
+		},
+		WaitState: NewWaitState(datastore.MenderStateInventoryUpdateRetryWait, ToSync),
+		from:      from,
+		err:       err,
+	}
+}
+
+func (fir *inventoryUpdateRetry) Cancel() bool {
+	return fir.WaitState.Cancel()
+}
+
+func (fir *inventoryUpdateRetry) Handle(ctx *StateContext, c Controller) (State, bool) {
+	if ctx.nextAttemptAt.After(time.Now()) {
+		remainingWaitDuration := time.Until(ctx.nextAttemptAt)
+		configuredInterval := c.GetRetryPollInterval()
+		if remainingWaitDuration.Seconds() > configuredInterval.Seconds() {
+			remainingWaitDuration = configuredInterval
+		}
+		log.Infof("Handle update inventory retry state: not the time yet; %ds/%v remaining.",
+			remainingWaitDuration/time.Second, time.Until(ctx.nextAttemptAt))
+		return fir.Wait(NewInventoryUpdateState(), fir, remainingWaitDuration, ctx.WakeupChan)
+	}
+	log.Infof("Handle update inventory retry state try: %d", ctx.inventoryUpdateAttempts)
+	err := c.InventoryRefresh()
+	if err != nil {
+		log.Warnf("Failed to refresh inventory: %v", err)
+		if errors.Cause(err) == errNoArtifactName {
+			return NewErrorState(NewTransientError(err)), false
+		}
+	} else {
+		return States.CheckWait, false
+	}
+
+	interval, err := client.GetExponentialBackoffTime(
+		ctx.inventoryUpdateAttempts,
+		c.GetRetryPollInterval(),
+		c.GetRetryPollCount(),
+	)
+	ctx.nextAttemptAt = time.Now().Add(interval)
+	if err != nil {
+		log.Infof("Handle update inventory retry state: failed to send inventory: %s", err.Error())
+		return States.CheckWait, false
+	}
+
+	ctx.inventoryUpdateAttempts++
+
+	configuredInterval := c.GetRetryPollInterval()
+	if interval.Seconds() > configuredInterval.Seconds() {
+		interval = configuredInterval
+	}
+	log.Infof("Wait %v before next inventory update attempt in %v",
+		interval, time.Until(ctx.nextAttemptAt))
+	return fir.Wait(NewInventoryUpdateState(), fir, interval, ctx.WakeupChan)
+}
+
 type checkWaitState struct {
 	baseState
 	WaitState
@@ -1156,74 +1241,44 @@ func (cw *checkWaitState) Cancel() bool {
 func (cw *checkWaitState) Handle(ctx *StateContext, c Controller) (State, bool) {
 
 	log.Debugf("Handle check wait state")
-
-	// calculate next interval
-	update := ctx.lastUpdateCheckAttempt.Add(c.GetUpdatePollInterval())
-	inventory := ctx.lastInventoryUpdateAttempt.Add(c.GetInventoryPollInterval())
-
-	// if we haven't sent inventory so far
+	// Inventory should be sent at first try
 	if ctx.lastInventoryUpdateAttempt.IsZero() {
-		inventory = ctx.lastInventoryUpdateAttempt
+		return States.InventoryUpdate, false
 	}
 
-	log.Debugf("Check wait state; next checks: (update: %v) (inventory: %v)",
-		update, inventory)
+	nextUpdateCheck := ctx.lastUpdateCheckAttempt.Add(c.GetUpdatePollInterval())
+	nextInventoryCheck := ctx.lastInventoryUpdateAttempt.Add(c.GetInventoryPollInterval())
 
-	next := struct {
-		when  time.Time
-		state State
-	}{
-		// assume update will be the next state
-		when:  update,
-		state: States.UpdateCheck,
+	if nextUpdateCheck.Before(nextInventoryCheck) {
+		return cw.Wait(
+			States.UpdateCheck,
+			cw,
+			time.Until(nextUpdateCheck),
+			ctx.WakeupChan,
+		)
 	}
 
-	if inventory.Before(update) {
-		next.when = inventory
-		next.state = States.InventoryUpdate
-	}
-
-	now := time.Now()
-	log.Debugf("Next check: %v:%v, (%v)", next.when, next.state, now)
-
-	// check if we should wait for the next state or we should return
-	// immediately
-	var wait time.Duration
-	if next.when.After(now) {
-		wait = next.when.Sub(now)
-	}
-
-	// (MEN-2195): Set the last update/inventory check time to now, as an error in an enter script will
-	// hinder these states from ever running, and thus causing an infinite loop if the script
-	// keeps returning the same error.
-	switch (next.state).(type) {
-	case *inventoryUpdateState:
-		if wait == 0 {
-			ctx.lastInventoryUpdateAttempt = now
-		} else {
-			ctx.lastInventoryUpdateAttempt = next.when
-		}
-	case *updateCheckState:
-		if wait == 0 {
-			ctx.lastUpdateCheckAttempt = now
-		} else {
-			ctx.lastUpdateCheckAttempt = next.when
-		}
-	}
-
-	if wait != 0 {
-		log.Debugf("Waiting %s for the next state", wait)
-		return cw.Wait(next.state, cw, wait, ctx.WakeupChan)
-	}
-
-	log.Debugf("Check wait returned: %v", next.state)
-	return next.state, false
+	return cw.Wait(
+		States.InventoryUpdate,
+		cw,
+		time.Until(nextInventoryCheck),
+		ctx.WakeupChan,
+	)
 }
 
 type inventoryUpdateState struct {
 	baseState
 }
 
+func NewInventoryUpdateState() State {
+	return &inventoryUpdateState{
+		baseState: baseState{
+			id: datastore.MenderStateInventoryUpdate,
+			t:  ToSync,
+		},
+	}
+}
+
 func (iu *inventoryUpdateState) Handle(ctx *StateContext, c Controller) (State, bool) {
 
 	err := c.InventoryRefresh()
@@ -1232,9 +1287,14 @@ func (iu *inventoryUpdateState) Handle(ctx *StateContext, c Controller) (State,
 		if errors.Cause(err) == errNoArtifactName {
 			return NewErrorState(NewTransientError(err)), false
 		}
+		return NewInventoryUpdateRetryState(iu, err), false
 	} else {
 		log.Debugf("Inventory refresh complete")
 	}
+
+	// Reset the counter when done with retrying
+	ctx.inventoryUpdateAttempts = 0
+
 	return States.CheckWait, false
 }
 
@@ -1259,7 +1319,7 @@ func NewErrorState(err menderError) State {
 
 func (e *errorState) Handle(ctx *StateContext, c Controller) (State, bool) {
 	// stop deployment logging
-	DeploymentLogger.Disable()
+	_ = DeploymentLogger.Disable()
 
 	log.Infof("Handling error state, current error: %v", e.cause.Error())
 	// decide if error is transient, exit for now
@@ -1419,7 +1479,7 @@ func (usr *updateStatusReportState) Handle(ctx *StateContext, c Controller) (Sta
 
 	// start deployment logging; no error checking
 	// we can do nothing here; either we will have the logs or not...
-	DeploymentLogger.Enable(usr.Update().ID)
+	_ = DeploymentLogger.Enable(usr.Update().ID)
 
 	log.Debug("Handling update status report state")
 
@@ -1454,7 +1514,7 @@ func (usr *updateStatusReportState) Handle(ctx *StateContext, c Controller) (Sta
 
 	log.Debug("Reporting complete")
 	// stop deployment logging as the update is completed at this point
-	DeploymentLogger.Disable()
+	_ = DeploymentLogger.Disable()
 
 	return States.Idle, false
 }
@@ -1539,6 +1599,35 @@ func (usr *updateStatusReportRetryState) PermitLooping() bool {
 	return true
 }
 
+type updateRebootPauseRequestedState struct {
+	*updateState
+}
+
+// This state solves a tricky issue. When checking Update Control Maps before
+// the reboot state, we want a few things to happen: If a pause is requested, we
+// want a spontaneous reboot to act as if the update resumed after a normal
+// reboot. This requires that we store datastore.MenderStateReboot in the
+// database, so that the recovery logic will work. However, we can not store
+// this *before* checking the Update Control Maps, because the action may
+// resolve to "fail", and then it would be very bad to resume the update. So the
+// maps need to be checked, and then, *only if* there is a pause, we enter the
+// state below to store the datastore.MenderStateReboot in the database. Then we
+// go back to pausing.
+func NewUpdateRebootPauseRequestedState(update *datastore.UpdateInfo) UpdateState {
+	return &updateRebootPauseRequestedState{
+		updateState: NewUpdateState(datastore.MenderStateReboot,
+			ToArtifactReboot_Enter, update),
+	}
+}
+
+func (s *updateRebootPauseRequestedState) Handle(ctx *StateContext, c Controller) (State, bool) {
+	// No need to fetch maps from the server. If we entered this state then
+	// it means we just fetched them. Go directly to checking the action
+	// instead. Note that we also pass nil as the pauseState. Since we are
+	// in this state, the next time around, we want a real pause to happen.
+	return NewControlMapState(NewUpdateRebootState(s.Update()), nil), false
+}
+
 type updateRebootState struct {
 	*updateState
 }
@@ -1635,19 +1724,19 @@ func (rs *updateAfterRebootState) Handle(ctx *StateContext,
 	c Controller) (State, bool) {
 	// start deployment logging; no error checking
 	// we can do nothing here; either we will have the logs or not...
-	DeploymentLogger.Enable(rs.Update().ID)
+	_ = DeploymentLogger.Enable(rs.Update().ID)
 
 	// this state is needed to satisfy ToReboot transition Leave() action
 	log.Debug("Handling state after reboot")
 
-	return NewControlMapState(NewUpdateCommitState(rs.Update())), false
+	return NewFetchControlMapState(NewUpdateCommitState(rs.Update()), nil), false
 }
 
 type updateRollbackState struct {
 	*updateState
 }
 
-func NewUpdateRollbackState(update *datastore.UpdateInfo) State {
+func NewUpdateRollbackState(update *datastore.UpdateInfo) *updateRollbackState {
 	return &updateRollbackState{
 		updateState: NewUpdateState(datastore.MenderStateRollback, ToArtifactRollback, update),
 	}
@@ -1669,6 +1758,11 @@ func (rs *updateRollbackState) Handle(ctx *StateContext, c Controller) (State, b
 			return rs.HandleError(ctx, c, NewFatalError(err))
 		}
 	}
+
+	// Query the rollback parameter from the Update Module, in case it has
+	// not been called previously (MEN-4882)
+	rs.queryUpdateModuleReboot(c)
+
 	for n := range c.GetInstallers() {
 		rebootRequested, err := rs.Update().RebootRequested.Get(n)
 		if err != nil {
@@ -1698,7 +1792,53 @@ func (rs *updateRollbackState) Handle(ctx *StateContext, c Controller) (State, b
 		rs.Update()), false
 }
 
-func (rs *updateRollbackState) HandleError(ctx *StateContext, c Controller, merr menderError) (State, bool) {
+func (rs *updateRollbackState) queryUpdateModuleReboot(c Controller) {
+	for n, i := range c.GetInstallers() {
+		v, err := rs.Update().RebootRequested.Get(n)
+		if err != nil {
+			log.Debugf(
+				"Failed to query the database for the Update Module's"+
+					" ArtifactRollback parameter during updateRollbackState: %s."+
+					"Will query the update module anew.",
+				err.Error())
+		}
+		switch v {
+		// If the value is 'RebootRequested', or 'RebootTypeAutomatic',
+		// do not risk writing a 'RebootTypeNone', reboot in any case
+		case datastore.RebootTypeAutomatic, datastore.RebootTypeCustom:
+			continue
+		default:
+			needsReboot, err := i.NeedsReboot()
+			if err != nil {
+				log.Error(err.Error())
+				continue
+			}
+			switch needsReboot {
+			case installer.NoReboot:
+				err = rs.Update().RebootRequested.Set(n, datastore.RebootTypeNone)
+			case installer.RebootRequired:
+				err = rs.Update().RebootRequested.Set(n, datastore.RebootTypeCustom)
+			case installer.AutomaticReboot:
+				err = rs.Update().RebootRequested.Set(n, datastore.RebootTypeAutomatic)
+			default:
+				log.Error("Unknown reboot value returned from Update Module")
+			}
+			if err != nil {
+				log.Errorf(
+					"Unable to set the value returned from the update module in the database."+
+						" Error: %s",
+					err.Error(),
+				)
+			}
+		}
+	}
+}
+
+func (rs *updateRollbackState) HandleError(
+	ctx *StateContext,
+	c Controller,
+	merr menderError,
+) (State, bool) {
 	log.Error(merr.Error())
 	setBrokenArtifactFlag(ctx.Store, rs.Update().ArtifactName())
 	return NewUpdateErrorState(merr, rs.Update()), false
@@ -1864,12 +2004,14 @@ func setBrokenArtifactFlag(store store.Store, artName string) {
 
 type controlMapState struct {
 	wrappedState UpdateState
+	pauseState   UpdateState
 	baseState
 }
 
-func NewControlMapState(wrapsState UpdateState) State {
+func NewControlMapState(wrapsState, pauseState UpdateState) State {
 	return &controlMapState{
 		wrappedState: wrapsState,
+		pauseState:   pauseState,
 		baseState: baseState{
 			id: datastore.MenderStateUpdateControl,
 			t:  ToNone,
@@ -1888,7 +2030,12 @@ func (c *controlMapState) pauseName(t Transition) string {
 	case ToArtifactInstall:
 		return "pause_before_installing"
 	}
-	panic(fmt.Sprintf("No pause name mapping for: %s. This is a logic error in the state machine code", t))
+	panic(
+		fmt.Sprintf(
+			"No pause name mapping for: %s. This is a logic error in the state machine code",
+			t,
+		),
+	)
 }
 
 func (c *controlMapState) mapStateToName(t Transition) string {
@@ -1898,19 +2045,30 @@ func (c *controlMapState) mapStateToName(t Transition) string {
 	case ToArtifactInstall:
 		return "ArtifactInstall_Enter"
 	}
-	panic(fmt.Sprintf("No state name mapping for: %s. This is a logic error in the state machine code", t))
+	panic(
+		fmt.Sprintf(
+			"No state name mapping for: %s. This is a logic error in the state machine code",
+			t,
+		),
+	)
 }
 
 func (c *controlMapState) Handle(ctx *StateContext, controller Controller) (State, bool) {
 
 	log.Debugf("Handling update control state")
 
-	action := controller.GetControlMapPool().QueryAndUpdate(c.mapStateToName(c.wrappedState.Transition()))
+	action := controller.GetControlMapPool().
+		QueryAndUpdate(c.mapStateToName(c.wrappedState.Transition()))
 	log.Debugf("controlMapState action: %s", action)
 	switch action {
 	case "continue":
 		return c.wrappedState, false
 	case "pause":
+		if c.pauseState != nil {
+			log.Debug("Going to pause state")
+			return c.pauseState, false
+		}
+
 		log.Infof("Update Control: Pausing before entering %s state", c.wrappedState.Id())
 		log.Debugf("Reporting update status: %s", c.pauseName(c.wrappedState.Transition()))
 		if !ctx.pauseReported[c.wrappedState.Id().String()] {
@@ -1936,19 +2094,20 @@ func (c *controlMapState) Handle(ctx *StateContext, controller Controller) (Stat
 
 type fetchControlMapState struct {
 	baseState
-	update       *datastore.UpdateInfo
 	wrappedState UpdateState
+	pauseState   UpdateState
 }
 
 func (c *fetchControlMapState) PermitLooping() bool { return true }
 
-func NewFetchControlMapState(wrappedState UpdateState) State {
+func NewFetchControlMapState(wrappedState, pauseState UpdateState) State {
 	return &fetchControlMapState{
 		baseState: baseState{
 			id: datastore.MenderStateFetchUpdateControl,
 			t:  ToNone,
 		},
 		wrappedState: wrappedState,
+		pauseState:   pauseState,
 	}
 }
 
@@ -1956,31 +2115,36 @@ func (c *fetchControlMapState) Handle(ctx *StateContext, controller Controller)
 
 	log.Debugf("Handle fetchControlMap state")
 
-	// The update control maps are magically updated in here
-	_, err := controller.CheckUpdate()
+	if controller.GetControlMapPool().HasControlMap(c.wrappedState.Update().ID) {
 
-	if err != nil {
-		if errors.Is(err, client.ErrNoDeploymentAvailable) {
-			return c.wrappedState.HandleError(ctx, controller,
-				NewTransientError(errors.New("The deployment was aborted from the server")))
+		err := controller.RefreshServerUpdateControlMap(c.wrappedState.Update().ID)
+		if err != nil {
+			if errors.Is(err, client.ErrNoDeploymentAvailable) {
+				return c.wrappedState.HandleError(ctx, controller,
+					NewTransientError(errors.New("The deployment was aborted from the server")))
+			}
+
+			log.Errorf("Update control map check failed: %s, retrying...", err.Error())
+			return NewFetchRetryControlMapState(c.wrappedState, c.pauseState), false
 		}
 
-		log.Errorf("Update control map check failed: %s, retrying...", err.Error())
-		return NewFetchRetryControlMapState(c.wrappedState), false
 	}
 
-	return NewControlMapState(c.wrappedState), false
+	// Reset the retry count
+	ctx.controlMapFetchAttempts = 0
+
+	return NewControlMapState(c.wrappedState, c.pauseState), false
 }
 
 type fetchRetryControlMapState struct {
 	waitState
 	wrappedState UpdateState
-	retries      int
+	pauseState   UpdateState
 }
 
 func (c *fetchRetryControlMapState) PermitLooping() bool { return true }
 
-func NewFetchRetryControlMapState(wrappedState UpdateState) State {
+func NewFetchRetryControlMapState(wrappedState, pauseState UpdateState) State {
 	return &fetchRetryControlMapState{
 		waitState: waitState{
 			baseState: baseState{
@@ -1989,6 +2153,7 @@ func NewFetchRetryControlMapState(wrappedState UpdateState) State {
 			},
 		},
 		wrappedState: wrappedState,
+		pauseState:   pauseState,
 	}
 }
 
@@ -1996,17 +2161,21 @@ func (f *fetchRetryControlMapState) Handle(ctx *StateContext, c Controller) (Sta
 
 	log.Debugf("Handle fetch update control retry state")
 
-	intvl, err := client.GetExponentialBackoffTime(f.retries, c.GetUpdatePollInterval())
+	intvl, err := client.GetExponentialBackoffTime(
+		ctx.controlMapFetchAttempts,
+		c.GetUpdatePollInterval(),
+		c.GetRetryPollCount(),
+	)
 	if err != nil {
-		return NewUpdateErrorState(
-			NewTransientError(err), f.wrappedState.Update()), false
+		return f.wrappedState.HandleError(ctx, c,
+			NewTransientError(err))
 	}
 
-	f.retries++
+	ctx.controlMapFetchAttempts++
 
 	log.Infof("Wait %v before next update control map fetch/update attempt", intvl)
 	return f.Wait(
-		NewFetchControlMapState(f.wrappedState),
+		NewFetchControlMapState(f.wrappedState, f.pauseState),
 		f,
 		intvl,
 		ctx.WakeupChan,
@@ -2053,7 +2222,7 @@ func (c *controlMapPauseState) Handle(ctx *StateContext, controller Controller)
 
 		if errors.Is(err, NoUpdateMapsErr) {
 			log.Error("No control maps no longer present, continuing")
-			return NewControlMapState(c.wrappedState), false
+			return NewControlMapState(c.wrappedState, nil), false
 		}
 
 		updateMapFromAnyIn := time.Until(nextMapRefresh)
@@ -2063,7 +2232,7 @@ func (c *controlMapPauseState) Handle(ctx *StateContext, controller Controller)
 		}
 
 		return c.Wait(
-			NewControlMapState(c.wrappedState),
+			NewControlMapState(c.wrappedState, nil),
 			c,
 			updateMapFromAnyIn,
 			controller.GetControlMapPool().Updates)
@@ -2071,7 +2240,13 @@ func (c *controlMapPauseState) Handle(ctx *StateContext, controller Controller)
 
 	updateMapFromServerIn := time.Until(nextServerMapRefresh)
 
-	if updateMapFromServerIn < 0 {
+	// For rapid feedback, use the smallest number of the map refresh time
+	// and the update poll interval. The latter is often much shorter.
+	if updateMapFromServerIn > controller.GetUpdatePollInterval() {
+		updateMapFromServerIn = controller.GetUpdatePollInterval()
+	}
+
+	if updateMapFromServerIn <= 0 {
 		updateMapFromServerIn = 30
 	}
 
@@ -2079,8 +2254,8 @@ func (c *controlMapPauseState) Handle(ctx *StateContext, controller Controller)
 
 	log.Debug("Pausing the event loop")
 	return c.MultiplexWait(
-		NewFetchControlMapState(c.wrappedState), // on ticker expiry
-		NewControlMapState(c.wrappedState),      // on map updates
+		NewFetchControlMapState(c.wrappedState, nil), // on ticker expiry
+		NewControlMapState(c.wrappedState, nil),      // on map updates
 		updateMapFromServerIn,
 		controller.GetControlMapPool().Updates,
 	)
@@ -2090,7 +2265,10 @@ type UpdateControlMapWaitState struct {
 	waitState
 }
 
-func NewUpdateControlMapWaitState(id datastore.MenderState, t Transition) *UpdateControlMapWaitState {
+func NewUpdateControlMapWaitState(
+	id datastore.MenderState,
+	t Transition,
+) *UpdateControlMapWaitState {
 	return &UpdateControlMapWaitState{
 		waitState{
 			baseState: baseState{id: id, t: t},
@@ -2100,25 +2278,25 @@ func NewUpdateControlMapWaitState(id datastore.MenderState, t Transition) *Updat
 }
 
 // MultiplexWait multiplexes the next state depending on the action which occurs
-// first. If the ticker expires, it goes to 'tickerState', and if a 'wakeup' is
+// first. If the timer expires, it goes to 'timerState', and if a 'wakeup' is
 // received, it goes to the 'wakeupState'.
 func (ws *UpdateControlMapWaitState) MultiplexWait(
-	tickerState, wakeupState State,
+	timerState, wakeupState State,
 	wait time.Duration,
 	wakeup chan bool) (State, bool) {
-	ticker := time.NewTicker(wait)
+	timer := timers.Get(wait)
 	ws.wakeup = wakeup
 
-	defer ticker.Stop()
+	defer timers.Put(timer)
 	select {
-	case <-ticker.C:
-		log.Debugf("Wait complete")
-		return tickerState, false
+	case <-timer.C:
+		log.Debug("Wait complete")
+		return timerState, false
 	case <-ws.wakeup:
 		log.Info("Forced wake-up from sleep")
 		return wakeupState, false
 	case <-ws.cancel:
-		log.Infof("Wait canceled")
+		log.Info("Wait canceled")
 	}
-	return tickerState, true
+	return timerState, true
 }
