@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -14,17 +14,21 @@
 package client
 
 import (
+	"context"
 	"encoding/json"
 	"fmt"
 	"io"
 	"net"
 	"net/http"
+	"net/url"
+	"os"
 	"path"
 	"strings"
 	"time"
 
 	"io/ioutil"
 
+	"github.com/gorilla/websocket"
 	"github.com/pkg/errors"
 	log "github.com/sirupsen/logrus"
 
@@ -65,6 +69,8 @@ var (
 
 	// connection keepalive options
 	connectionKeepaliveTime = 10 * time.Second
+
+	ErrClientUnauthorized = errors.New("Client is unauthorized")
 )
 
 // Mender API Client wrapper. A standard http.Client is compatible with this
@@ -76,6 +82,13 @@ type ApiRequester interface {
 	Do(req *http.Request) (*http.Response, error)
 }
 
+// An ApiRequester which internally authorizes automatically. It's possible to
+// call ClearAuthorization() in order to force reauthorization.
+type AuthorizedApiRequester interface {
+	ApiRequester
+	ClearAuthorization()
+}
+
 // MenderServer is a placeholder for a full server definition used when
 // multiple servers are given. The fields corresponds to the definitions
 // given in MenderConfig.
@@ -108,7 +121,11 @@ func NewAPIError(err error, resp *http.Response) *APIError {
 
 func (a *APIError) Error() string {
 
-	err := fmt.Sprintf("(request_id: %s): %s", a.reqID, a.error.Error())
+	err := a.error.Error()
+
+	if a.reqID != "" {
+		err = fmt.Sprintf("(request_id: %s): %s", a.reqID, err)
+	}
 
 	if a.serverErrMsg != "" {
 		return err + fmt.Sprintf(" server error message: %s", a.serverErrMsg)
@@ -128,150 +145,130 @@ func (a *APIError) Unwrap() error {
 	return a.error
 }
 
-type RequestProcessingFunc func(response *http.Response) (interface{}, error)
-
-// wrapper for http.Client with additional methods
 type ApiClient struct {
 	http.Client
 }
 
-// function type for reauthorization closure (see func reauthorize@mender.go)
-type ClientReauthorizeFunc func(string) (AuthToken, error)
-
-// function type for setting server (in case of multiple fallover servers)
-type ServerManagementFunc func() *MenderServer
-
-// Return a new ApiRequest
-func (a *ApiClient) Request(code AuthToken, nextServerIterator ServerManagementFunc, reauth ClientReauthorizeFunc) *ApiRequest {
-	return &ApiRequest{
-		api:                a,
-		auth:               code,
-		nextServerIterator: nextServerIterator,
-		revoke:             reauth,
-	}
-}
-
-// ApiRequester compatible helper. The helper can be used for executing API
-// requests that require authorization as provided Do() method will automatically
-// setup authorization information in the request.
-type ApiRequest struct {
-	api *ApiClient
+type ReauthorizingClient struct {
+	ApiClient
 	// authorization code to use for requests
 	auth AuthToken
+	// server to use for requests
+	serverURL ServerURL
 	// anonymous function to initiate reauthorization
 	revoke ClientReauthorizeFunc
-	// anonymous function to set server
-	nextServerIterator ServerManagementFunc
 }
 
-// tryDo is a wrapper around http.Do that also tries to reauthorize
-// on a 401 response (Unauthorized).
-func (ar *ApiRequest) tryDo(req *http.Request, serverURL string) (*http.Response, error) {
-	r, err := ar.api.Do(req)
-	if err == nil && r.StatusCode == http.StatusUnauthorized {
-		// invalid JWT; most likely the token is expired:
-		// Try to refresh it and reattempt sending the request
-		log.Info("Device unauthorized; attempting reauthorization")
-		if jwt, e := ar.revoke(serverURL); e == nil {
-			// retry API request with new JWT token
-			ar.auth = jwt
-			// check if request had a body
-			// (GetBody is optional, and nil if body is empty)
-			if req.GetBody != nil {
-				if body, e := req.GetBody(); e == nil {
-					req.Body = body
-				}
-			}
-			req.Header.Set("Authorization", fmt.Sprintf("Bearer %s", ar.auth))
-			r, err = ar.api.Do(req)
-		} else {
-			log.Warnf("Reauthorization failed with error: %s", e.Error())
-		}
+// function type for reauthorization closure (see func reauthorize@mender.go)
+type ClientReauthorizeFunc func() (AuthToken, ServerURL, error)
+
+// Reconstruct the given request, taking JWT token and serverURL into account.
+func (c *ReauthorizingClient) reconstructRequest(req *http.Request) (*http.Request, error) {
+	if c.serverURL == "" {
+		return nil, ErrClientUnauthorized
 	}
-	return r, err
+
+	serverURL, err := url.Parse(string(c.serverURL))
+	if err != nil {
+		return nil, errors.Wrap(err, "Could not parse ServerURL from auth manager")
+	}
+
+	// First prefill newURL with existing URL.
+	newURL, _ := url.Parse(req.URL.String())
+
+	// Then selectively fill in the prefix from ServerURL.
+	newURL.Scheme = serverURL.Scheme
+	newURL.Host = serverURL.Host
+	newURL.Path = strings.TrimRight(serverURL.Path, "/") +
+		"/" +
+		strings.TrimLeft(req.URL.Path, "/")
+
+	log.Debugf("Connecting to server %s", serverURL.String())
+
+	var body io.ReadCloser
+	if req.GetBody != nil {
+		body, err = req.GetBody()
+		if err != nil {
+			return nil, errors.Wrap(err, "Unable to reconstruct HTTP request body")
+		}
+	} else {
+		body = nil
+	}
+
+	// create a new request object to avoid issues when consuming
+	// the request body multiple times when failing over a different
+	// server. It is not safe to reuse the same request multiple times
+	// when request.Body is not nil
+	//
+	// see: https://github.com/golang/go/issues/19653
+	// Error message: http: ContentLength=52 with Body length 0
+	newReq, _ := http.NewRequest(req.Method, newURL.String(), body)
+	newReq.Header = req.Header
+	newReq.GetBody = req.GetBody
+	newReq.Header.Set("Authorization", fmt.Sprintf("Bearer %s", c.auth))
+
+	return newReq, nil
 }
 
 // Do is a wrapper for http.Do function for ApiRequests. This function in
 // addition to calling http.Do handles client-server authorization header /
 // reauthorization, as well as attempting failover servers (if given) whenever
 // the server "refuse" to serve the request.
-func (ar *ApiRequest) Do(req *http.Request) (*http.Response, error) {
-	if ar.nextServerIterator == nil {
-		return nil, errors.New("Empty server list!")
-	}
-	if req.Header.Get("Authorization") == "" {
-		// Add JWT to header
-		req.Header.Set("Authorization", fmt.Sprintf("Bearer %s", ar.auth))
-	}
-	var r *http.Response
-	var host string
-	var err error
-
-	server := ar.nextServerIterator()
+func (c *ReauthorizingClient) Do(req *http.Request) (*http.Response, error) {
+	fetchedNewToken := false
 	for {
-		// Split host from URL
-		tmp := strings.Split(server.ServerURL, "://")
-		if len(tmp) == 1 {
-			host = tmp[0]
-		} else {
-			// (len >= 2) should usually be 2
-			host = tmp[1]
+		var r *http.Response
+		newReq, err := c.reconstructRequest(req)
+		if err == nil {
+			r, err = c.ApiClient.Do(newReq)
+		} else if err != ErrClientUnauthorized {
+			return nil, err
 		}
 
-		req.URL.Host = host
-		req.Host = host
-
-		log.Debugf("Connecting to server %s", host)
-
-		var body io.ReadCloser
-		if req.GetBody != nil {
-			body, err = req.GetBody()
-			if err != nil {
-				return nil, errors.Wrap(err, "Unable to reconstruct HTTP request body")
+		// If we haven't yet fetched a new token, and the call results
+		// in connection error or Unauthorized status, try to get
+		// one. Otherwise return the result as is.
+		if !fetchedNewToken && (err != nil || r.StatusCode == http.StatusUnauthorized) {
+			// Try to fetch a new JWT token from one of the servers in the server
+			// list.
+			log.Info("Device unauthorized; attempting reauthorization")
+			jwt, serverURL, e := c.revoke()
+			if e == nil {
+				// retry API request with new JWT token
+				c.auth = jwt
+				c.serverURL = serverURL
+				log.Info("Reauthorization successful")
+			} else {
+				log.Warnf("Reauthorization failed with error: %s", e.Error())
+				return nil, e
 			}
+			fetchedNewToken = true
 		} else {
-			body = nil
-		}
-
-		// create a new request object to avoid issues when consuming
-		// the request body multiple times when failing over a different
-		// server. It is not safe to reuse the same request multiple times
-		// when request.Body is not nil
-		//
-		// see: https://github.com/golang/go/issues/19653
-		// Error message: http: ContentLength=52 with Body length 0
-		newReq, _ := http.NewRequest(req.Method, req.URL.String(), body)
-		newReq.Header = req.Header
-		newReq.GetBody = req.GetBody
-
-		r, err = ar.tryDo(newReq, server.ServerURL)
-		if err == nil && r.StatusCode < 400 {
-			break
-		}
-		prewHost := server.ServerURL
-		if server = ar.nextServerIterator(); server == nil {
-			break
-		}
-		log.Warnf("Server %q failed to serve request %q. Attempting %q",
-			prewHost, req.URL.Path, server.ServerURL)
-	}
-	if server != nil {
-		// reset server iterator
-		for {
-			if ar.nextServerIterator() == nil {
-				break
-			}
+			return r, err
 		}
 	}
-	return r, err
+}
+
+func (c *ReauthorizingClient) ClearAuthorization() {
+	c.auth = ""
+	c.serverURL = ""
+}
+
+func NewReauthorizingClient(
+	conf Config,
+	reauth ClientReauthorizeFunc,
+) (*ReauthorizingClient, error) {
+	client, err := NewApiClient(conf)
+	if err != nil {
+		return nil, err
+	}
+	return &ReauthorizingClient{
+		ApiClient: *client,
+		revoke:    reauth,
+	}, nil
 }
 
 func NewApiClient(conf Config) (*ApiClient, error) {
-	return New(conf)
-}
-
-// New initializes new client
-func New(conf Config) (*ApiClient, error) {
 
 	var client *http.Client
 	if conf == (Config{}) {
@@ -316,12 +313,30 @@ func nrOfSystemCertsFound(certDir string) (int, error) {
 	sysCertsFound := 0
 	files, err := ioutil.ReadDir(certDir)
 	if err != nil {
-		return 0, fmt.Errorf("Failed to read the OpenSSL default directory (%s). Err %v", certDir, err.Error())
+		return 0, fmt.Errorf(
+			"Failed to read the OpenSSL default directory (%s). Err %v",
+			certDir,
+			err.Error(),
+		)
 	}
 	for _, certFile := range files {
+		// Need to re-stat here because ReadDir does not resolve
+		// symlinks.
+		info, err := os.Stat(path.Join(certDir, certFile.Name()))
+		if err != nil {
+			log.Debugf("Failed to stat file %s: %s", certFile.Name(), err.Error())
+			continue
+		} else if !info.Mode().IsRegular() {
+			log.Debugf("Not a regular file, skipping: %s", info.Name())
+			continue
+		}
+
 		certBytes, err := ioutil.ReadFile(path.Join(certDir, certFile.Name()))
 		if err != nil {
-			log.Debugf("Failed to read the certificate file for the HttpsClient. Err %v", err.Error())
+			log.Debugf(
+				"Failed to read the certificate file for the HttpsClient. Err %v",
+				err.Error(),
+			)
 			continue
 		}
 
@@ -330,12 +345,13 @@ func nrOfSystemCertsFound(certDir string) (int, error) {
 			log.Debugf("No PEM certificate found in '%s'", certFile)
 			continue
 		}
-		first, certs := certs[0], certs[1:]
-		_, err = openssl.LoadCertificateFromPEM(first)
-		if err != nil {
-			log.Debug(err.Error())
-		} else {
-			sysCertsFound += 1
+		for _, cert := range certs {
+			_, err = openssl.LoadCertificateFromPEM(cert)
+			if err != nil {
+				log.Debug(err.Error())
+			} else {
+				sysCertsFound += 1
+			}
 		}
 	}
 	return sysCertsFound, nil
@@ -344,7 +360,11 @@ func nrOfSystemCertsFound(certDir string) (int, error) {
 func loadServerTrust(ctx *openssl.Ctx, conf *Config) (*openssl.Ctx, error) {
 	defaultCertDir, err := openssl.GetDefaultCertificateDirectory()
 	if err != nil {
-		return ctx, errors.Wrap(err, "Failed to get the default OpenSSL certificate directory. Please verify the OpenSSL setup")
+		return ctx, errors.Wrap(
+			err,
+			"Failed to get the default OpenSSL certificate directory. Please verify the"+
+				" OpenSSL setup",
+		)
 	}
 	sysCertsFound, err := nrOfSystemCertsFound(defaultCertDir)
 	if err != nil {
@@ -354,21 +374,28 @@ func loadServerTrust(ctx *openssl.Ctx, conf *Config) (*openssl.Ctx, error) {
 	// Set the default system certificate path for this OpenSSL context
 	err = ctx.SetDefaultVerifyPaths()
 	if err != nil {
-		return ctx, fmt.Errorf("Failed to set the default OpenSSL directory. OpenSSL error code: %s", err.Error())
+		return ctx, fmt.Errorf(
+			"Failed to set the default OpenSSL directory. OpenSSL error code: %s",
+			err.Error(),
+		)
 	}
-	// Load the server certificate into the OpenSSL context
-	err = ctx.LoadVerifyLocations(conf.ServerCert, "")
-	if err != nil {
-		if strings.Contains(err.Error(), "No such file or directory") {
-			log.Warnf(errMissingServerCertF, conf.ServerCert)
-		} else {
-			log.Errorf("Failed to Load the Server certificate. Err %s", err.Error())
-		}
-		// If no system certificates, nor a server certificate is found,
-		// warn the user, as this is a pretty common error.
-		if sysCertsFound == 0 {
-			log.Error(errMissingCerts)
+	if conf.ServerCert != "" {
+		// Load the server certificate into the OpenSSL context
+		err = ctx.LoadVerifyLocations(conf.ServerCert, "")
+		if err != nil {
+			if strings.Contains(err.Error(), "No such file or directory") {
+				log.Warnf(errMissingServerCertF, conf.ServerCert)
+			} else {
+				log.Errorf("Failed to Load the Server certificate. Err %s", err.Error())
+			}
+			// If no system certificates, nor a server certificate is found,
+			// warn the user, as this is a pretty common error.
+			if sysCertsFound == 0 {
+				log.Error(errMissingCerts)
+			}
 		}
+	} else if sysCertsFound == 0 {
+		log.Warn(errMissingCerts)
 	}
 	return ctx, err
 }
@@ -392,7 +419,8 @@ func loadPrivateKey(keyFile string, engineId string) (key openssl.PrivateKey, er
 	} else {
 		keyBytes, err := ioutil.ReadFile(keyFile)
 		if err != nil {
-			return nil, errors.Wrap(err, "Private key file from the HttpsClient configuration not found")
+			return nil, errors.Wrap(err, "Private key file from the HttpsClient configuration"+
+				" not found")
 		}
 
 		key, err = openssl.LoadPrivateKeyFromPEM(keyBytes)
@@ -456,8 +484,7 @@ func loadClientTrust(ctx *openssl.Ctx, conf *Config) (*openssl.Ctx, error) {
 	return ctx, nil
 }
 
-func dialOpenSSL(ctx *openssl.Ctx, conf *Config, network string, addr string) (net.Conn, error) {
-
+func dialOpenSSL(ctx *openssl.Ctx, conf *Config, _ string, addr string) (net.Conn, error) {
 	flags := openssl.DialFlags(0)
 
 	if conf.NoVerify {
@@ -496,9 +523,7 @@ func dialOpenSSL(ctx *openssl.Ctx, conf *Config, network string, addr string) (n
 	return conn, err
 }
 
-func newHttpsClient(conf Config) (*http.Client, error) {
-	client := newHttpClient()
-
+func newOpenSSLCtx(conf Config) (*openssl.Ctx, error) {
 	ctx, err := openssl.NewCtx()
 	if err != nil {
 		return nil, err
@@ -519,13 +544,32 @@ func newHttpsClient(conf Config) (*http.Client, error) {
 	if conf.NoVerify {
 		log.Warn("certificate verification skipped..")
 	}
+
+	return ctx, nil
+}
+
+func newHttpsClient(conf Config) (*http.Client, error) {
+	ctx, err := newOpenSSLCtx(conf)
+	if err != nil {
+		return nil, err
+	}
+
+	disableKeepAlive := false
+	idleConnTimeoutSeconds := 0
+	if conf.Connectivity != nil {
+		disableKeepAlive = conf.Connectivity.DisableKeepAlive
+		idleConnTimeoutSeconds = conf.Connectivity.IdleConnTimeoutSeconds
+	}
 	transport := http.Transport{
-		Proxy: http.ProxyFromEnvironment,
+		DisableKeepAlives: disableKeepAlive,
+		IdleConnTimeout:   time.Duration(idleConnTimeoutSeconds) * time.Second,
+		Proxy:             http.ProxyFromEnvironment,
 		DialTLS: func(network string, addr string) (net.Conn, error) {
 			return dialOpenSSL(ctx, &conf, network, addr)
 		},
 	}
 
+	client := newHttpClient()
 	client.Transport = &transport
 	return client, nil
 }
@@ -551,18 +595,38 @@ type Security struct {
 	SSLEngine      string `json:",omitempty"`
 }
 
+// Connectivity instructs the client how we want to treat the keep alive connections
+// and when a connection is considered idle and therefore closed
+// NOTE: Careful when changing this, the struct is exposed directly in the
+// 'mender.conf' file.
+type Connectivity struct {
+	// If set to true, there will be no persistent connections, and every
+	// HTTP transaction will try to establish a new connection
+	DisableKeepAlive bool `json:",omitempty"`
+	// A number of seconds after which a connection is considered idle and closed.
+	// The longer this is the longer connections are up after the first call over HTTP
+	IdleConnTimeoutSeconds int `json:",omitempty"`
+}
+
 func (h *HttpsClient) Validate() {
 	if h == nil {
 		return
 	}
 	if h.Certificate != "" || h.Key != "" {
 		if h.Certificate == "" {
-			log.Error("The 'Key' field is set in the mTLS configuration, but no 'Certificate' is given. Both need to be present in order for mTLS to function")
+			log.Error(
+				"The 'Key' field is set in the mTLS configuration, but no 'Certificate' is given." +
+					" Both need to be present in order for mTLS to function",
+			)
 		}
 		if h.Key == "" {
-			log.Error("The 'Certificate' field is set in the mTLS configuration, but no 'Key' is given. Both need to be present in order for mTLS to function")
+			log.Error(
+				"The 'Certificate' field is set in the mTLS configuration, but no 'Key' is given." +
+					" Both need to be present in order for mTLS to function",
+			)
 		} else if strings.HasPrefix(h.Key, pkcs11URIPrefix) && len(h.SSLEngine) == 0 {
-			log.Errorf("The 'Key' field is set to be loaded from %s, but no 'SSLEngine' is given. Both need to be present in order for loading of the key to function",
+			log.Errorf("The 'Key' field is set to be loaded from %s, but no 'SSLEngine' is given."+
+				" Both need to be present in order for loading of the key to function",
 				pkcs11URIPrefix)
 		}
 	}
@@ -571,6 +635,7 @@ func (h *HttpsClient) Validate() {
 type Config struct {
 	ServerCert string
 	*HttpsClient
+	*Connectivity
 	NoVerify bool
 }
 
@@ -594,17 +659,23 @@ var MaxRetriesExceededError = errors.New("Tried maximum amount of times")
 // Simple algorithm: Start with one minute, and try three times, then double
 // interval (maxInterval is maximum) and try again. Repeat until we tried
 // three times with maxInterval.
-func GetExponentialBackoffTime(tried int, maxInterval time.Duration) (time.Duration, error) {
+func GetExponentialBackoffTime(tried int,
+	maxInterval time.Duration,
+	maxAttempts int) (time.Duration, error) {
 	const perIntervalAttempts = 3
 
 	interval := 1 * ExponentialBackoffSmallestUnit
 	nextInterval := interval
 
+	if maxAttempts > 0 && tried >= maxAttempts {
+		return 0, MaxRetriesExceededError
+	}
+
 	for c := 0; c <= tried; c += perIntervalAttempts {
 		interval = nextInterval
 		nextInterval *= 2
 		if interval >= maxInterval {
-			if tried-c >= perIntervalAttempts {
+			if maxAttempts <= 0 && tried-c >= perIntervalAttempts {
 				// At max interval and already tried three
 				// times. Give up.
 				return 0, MaxRetriesExceededError
@@ -637,3 +708,33 @@ func unmarshalErrorMessage(r io.Reader) string {
 	}
 	return e.Error
 }
+
+func newWebsocketDialerTLS(conf Config) (*websocket.Dialer, error) {
+	ctx, err := newOpenSSLCtx(conf)
+	if err != nil {
+		return nil, err
+	}
+
+	dialer := websocket.Dialer{
+		NetDialTLSContext: func(_ context.Context, network string, addr string) (net.Conn, error) {
+			return dialOpenSSL(ctx, &conf, network, addr)
+		},
+	}
+
+	return &dialer, nil
+}
+
+func NewWebsocketDialer(conf Config) (*websocket.Dialer, error) {
+	var dialer *websocket.Dialer
+	var err error
+	if conf == (Config{}) {
+		dialer = websocket.DefaultDialer
+	} else {
+		dialer, err = newWebsocketDialerTLS(conf)
+	}
+	if err != nil {
+		return nil, err
+	}
+
+	return dialer, nil
+}
