@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -500,8 +500,11 @@ func TestMakeUpdateCheckRequest(t *testing.T) {
 	assert.NotNil(t, reqs[getV1])
 	assert.NoError(t, err)
 
-	assert.Equal(t, "http://foo.bar/api/devices/v1/deployments/device/deployments/next?artifact_name=release-1",
-		reqs[getV1].URL.String())
+	assert.Equal(
+		t,
+		"http://foo.bar/api/devices/v1/deployments/device/deployments/next?artifact_name=release-1",
+		reqs[getV1].URL.String(),
+	)
 	assert.Equal(t, "http://foo.bar/api/devices/v1/deployments/device/deployments/next",
 		reqs[postV1].URL.String())
 	assert.Equal(t, "http://foo.bar/api/devices/v2/deployments/device/deployments/next",
@@ -514,7 +517,12 @@ func TestMakeUpdateCheckRequest(t *testing.T) {
 	require.Contains(t, params, "device_provides")
 	require.IsType(t, map[string]interface{}{}, params["device_provides"], string(body))
 	require.Contains(t, params["device_provides"], "artifact_name", string(body))
-	assert.Equal(t, "release-1", params["device_provides"].(map[string]interface{})["artifact_name"], string(body))
+	assert.Equal(
+		t,
+		"release-1",
+		params["device_provides"].(map[string]interface{})["artifact_name"],
+		string(body),
+	)
 	assert.Equal(t, true, params["update_control_map"])
 	body, err = ioutil.ReadAll(reqs[postV1].Body)
 	assert.NoError(t, err)
@@ -533,8 +541,11 @@ func TestMakeUpdateCheckRequest(t *testing.T) {
 	assert.NotNil(t, reqs[getV1])
 	assert.NoError(t, err)
 
-	assert.Equal(t, "http://foo.bar/api/devices/v1/deployments/device/deployments/next?artifact_name=foo&device_type=hammer",
-		reqs[getV1].URL.String())
+	assert.Equal(
+		t,
+		"http://foo.bar/api/devices/v1/deployments/device/deployments/next?artifact_name=foo&device_type=hammer",
+		reqs[getV1].URL.String(),
+	)
 	assert.Equal(t, "http://foo.bar/api/devices/v1/deployments/device/deployments/next",
 		reqs[postV1].URL.String())
 	assert.Equal(t, "http://foo.bar/api/devices/v2/deployments/device/deployments/next",
@@ -547,8 +558,18 @@ func TestMakeUpdateCheckRequest(t *testing.T) {
 	require.Contains(t, params, "device_provides")
 	require.IsType(t, map[string]interface{}{}, params["device_provides"], string(body))
 	require.Contains(t, params["device_provides"], "artifact_name", string(body))
-	assert.Equal(t, "foo", params["device_provides"].(map[string]interface{})["artifact_name"], string(body))
-	assert.Equal(t, "hammer", params["device_provides"].(map[string]interface{})["device_type"], string(body))
+	assert.Equal(
+		t,
+		"foo",
+		params["device_provides"].(map[string]interface{})["artifact_name"],
+		string(body),
+	)
+	assert.Equal(
+		t,
+		"hammer",
+		params["device_provides"].(map[string]interface{})["device_type"],
+		string(body),
+	)
 	assert.Equal(t, true, params["update_control_map"])
 	body, err = ioutil.ReadAll(reqs[postV1].Body)
 	assert.NoError(t, err)
@@ -690,3 +711,67 @@ func TestGetUpdateInfo(t *testing.T) {
 		})
 	}
 }
+
+func TestGetUpdateControlMap(t *testing.T) {
+	tests := map[string]struct {
+		httpHandler http.Handler
+		assertFunc  assert.ErrorAssertionFunc
+	}{
+		"OK": {
+			httpHandler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				w.WriteHeader(http.StatusOK)
+				w.Header().Set("Content-Type", "application/json")
+				w.Write([]byte(`{
+                                    "update_control_map": {
+					"ID": "3380e4f2-c913-11eb-9119-c39aba66b261",
+					"Priority": 1,
+					"States": {
+						"ArtifactInstall_Enter": {
+							"Action": "pause"
+						}
+					}
+                                    }
+				}`))
+			}),
+			assertFunc: assert.NoError,
+		},
+		"Error: wrong map format": {
+			httpHandler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				w.WriteHeader(http.StatusOK)
+				w.Header().Set("Content-Type", "application/json")
+				w.Write([]byte(`{}`))
+			}),
+			assertFunc: assert.Error,
+		},
+		"Error - BadRequest": {
+			httpHandler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				w.WriteHeader(http.StatusBadRequest)
+			}),
+			assertFunc: assert.Error,
+		},
+		"Error - no map on the server": {
+			httpHandler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				w.WriteHeader(http.StatusNotFound)
+			}),
+			assertFunc: assert.Error,
+		},
+		"Error - Internal server errror": {
+			httpHandler: http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				w.WriteHeader(http.StatusInternalServerError)
+			}),
+			assertFunc: assert.Error,
+		},
+	}
+
+	for name, test := range tests {
+		ts := httptest.NewUnstartedServer(
+			test.httpHandler,
+		)
+		t.Logf("Running test: %s\n", name)
+		ts.Start()
+		tc := ts.Client()
+		_, err := GetUpdateControlMap(tc, ts.URL, "")
+		test.assertFunc(t, err, "error: %v", err)
+		ts.Close()
+	}
+}
