@@ -1,4 +1,4 @@
-// Copyright 2021 Northern.tech AS
+// Copyright 2022 Northern.tech AS
 //
 //    Licensed under the Apache License, Version 2.0 (the "License");
 //    you may not use this file except in compliance with the License.
@@ -27,11 +27,12 @@ import (
 	"syscall"
 	"time"
 
+	"github.com/pkg/errors"
+	log "github.com/sirupsen/logrus"
+
 	"github.com/mendersoftware/mender-artifact/artifact"
 	"github.com/mendersoftware/mender-artifact/handlers"
 	"github.com/mendersoftware/mender/system"
-	"github.com/pkg/errors"
-	log "github.com/sirupsen/logrus"
 )
 
 type ModuleInstaller struct {
@@ -92,10 +93,11 @@ func (mod *ModuleInstaller) callModule(state string, capture bool) (string, erro
 	cmd := system.Command(mod.programPath, state, payloadPath)
 	cmd.Dir = mod.payloadPath()
 
-	stdoutLogger := newReadLogger(capture)
-	stderrLogger := newReadLogger(false)
-	cmd.Stdout = stdoutLogger
-	cmd.Stderr = stderrLogger
+	var buf *bytes.Buffer
+	if capture {
+		buf = bytes.NewBuffer(nil)
+		cmd.Stdout = buf
+	}
 	// Create new process group so we can kill them all instead of just the parent.
 	cmd.SysProcAttr = &syscall.SysProcAttr{
 		Setpgid: true,
@@ -118,37 +120,11 @@ func (mod *ModuleInstaller) callModule(state string, capture bool) (string, erro
 		log.Error(err.Error())
 	}
 
-	return stdoutLogger.output, err
-}
-
-type ReadLogger struct {
-	output  string
-	capture bool
-}
-
-func newReadLogger(capture bool) *ReadLogger {
-	return &ReadLogger{
-		output:  "",
-		capture: capture,
+	output := ""
+	if capture {
+		output = strings.TrimSuffix(buf.String(), "\n")
 	}
-}
-
-func (rl *ReadLogger) Write(p []byte) (int, error) {
-
-	buf := string(p)
-	buf = strings.TrimRight(buf, "\n")
-
-	if rl.capture {
-		for _, line := range strings.Split(buf, "\n") {
-			log.Debugf("Update module output: %s", line)
-		}
-		rl.output = rl.output + buf
-	} else {
-		for _, line := range strings.Split(buf, "\n") {
-			log.Infof("Update module output: %s", line)
-		}
-	}
-	return len(p), nil
+	return output, err
 }
 
 func (mod *ModuleInstaller) payloadPath() string {
@@ -210,43 +186,43 @@ func (mod *ModuleInstaller) buildStreamsTree(artifactHeaders,
 	}
 
 	filesAndContent := []fileNameAndContent{
-		fileNameAndContent{
+		{
 			"version",
 			fmt.Sprintf("%d", payloadHeaders.GetVersion()),
 		},
-		fileNameAndContent{
+		{
 			"current_artifact_name",
 			currName,
 		},
-		fileNameAndContent{
+		{
 			"current_artifact_group",
 			currGroup,
 		},
-		fileNameAndContent{
+		{
 			"current_device_type",
 			deviceType,
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "artifact_group"),
 			provides.ArtifactGroup,
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "artifact_name"),
 			provides.ArtifactName,
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "payload_type"),
 			mod.updateType,
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "header-info"),
 			string(headerInfoJson),
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "type-info"),
 			string(typeInfo),
 		},
-		fileNameAndContent{
+		{
 			path.Join("header", "meta-data"),
 			string(metaData),
 		},
@@ -529,7 +505,11 @@ func (d *moduleDownload) handleStreamChannel(err error) error {
 func (d *moduleDownload) handleFinishChannel() error {
 	d.finishFlag = true
 
-	if d.downloaderType != menderDownloader {
+	if d.downloaderType == menderDownloader {
+		// Make sure the file tree is synced and will not vanish across
+		// a reboot.
+		syscall.Sync()
+	} else {
 		// Publish empty entry to signal end of streams.
 		var err error
 		d.streamNext, err = d.publishNameInStreamNext("")
@@ -699,10 +679,6 @@ func (mod *ModuleInstaller) PrepareStoreUpdate() error {
 	storeUpdateCmd := system.Command(mod.programPath, "Download", payloadPath)
 	storeUpdateCmd.Dir = mod.payloadPath()
 
-	stdoutLogger := newReadLogger(false)
-	stderrLogger := newReadLogger(false)
-	storeUpdateCmd.Stdout = stdoutLogger
-	storeUpdateCmd.Stderr = stderrLogger
 	// Create new process group so we can kill them all instead of just the parent.
 	storeUpdateCmd.SysProcAttr = &syscall.SysProcAttr{
 		Setpgid: true,
@@ -771,8 +747,10 @@ func (mod *ModuleInstaller) NeedsReboot() (RebootAction, error) {
 		log.Debug("Module needs host reboot")
 		return AutomaticReboot, nil
 	} else {
-		return NoReboot, fmt.Errorf("Unexpected reply from update module NeedsArtifactReboot query: %s",
-			output)
+		return NoReboot, fmt.Errorf(
+			"Unexpected reply from update module NeedsArtifactReboot query: %s",
+			output,
+		)
 	}
 }
 
@@ -890,7 +868,10 @@ func NewModuleInstallerFactory(modulesPath, modulesWorkPath string,
 	}
 }
 
-func (mf *ModuleInstallerFactory) NewUpdateStorer(updateType string, payloadNum int) (handlers.UpdateStorer, error) {
+func (mf *ModuleInstallerFactory) NewUpdateStorer(
+	updateType *string,
+	payloadNum int,
+) (handlers.UpdateStorer, error) {
 	if payloadNum < 0 || payloadNum > 9999 {
 		return nil, fmt.Errorf("Payload index out of range 0-9999: %d", payloadNum)
 	}
@@ -899,8 +880,8 @@ func (mf *ModuleInstallerFactory) NewUpdateStorer(updateType string, payloadNum
 		payloadIndex:      payloadNum,
 		modulesPath:       mf.modulesPath,
 		modulesWorkPath:   mf.modulesWorkPath,
-		updateType:        updateType,
-		programPath:       path.Join(mf.modulesPath, updateType),
+		updateType:        *updateType,
+		programPath:       path.Join(mf.modulesPath, *updateType),
 		artifactInfo:      mf.artifactInfo,
 		deviceInfo:        mf.deviceInfo,
 		moduleTimeoutSecs: mf.moduleTimeoutSecs,
@@ -911,8 +892,12 @@ func (mf *ModuleInstallerFactory) NewUpdateStorer(updateType string, payloadNum
 func (mf *ModuleInstallerFactory) GetModuleTypes() []string {
 	fileList, err := ioutil.ReadDir(mf.modulesPath)
 	if err != nil {
-		log.Infof("Update Module path \"%s\" could not be opened (%s). Update modules will not be available",
-			mf.modulesPath, err.Error())
+		log.Infof(
+			"Update Module path \"%s\" could not be opened (%s)."+
+				" Update modules will not be available",
+			mf.modulesPath,
+			err.Error(),
+		)
 		return []string{}
 	}
 
